# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Author: Tuan Chien, Richard Hosking, Keegan Smith

import os
from datetime import timedelta
from typing import List, Optional, Tuple, Union
import re
from concurrent.futures import ThreadPoolExecutor, as_completed

import pendulum
from google.cloud.bigquery import SourceFormat, Client
from ratelimit import limits, sleep_and_retry
from tenacity import wait_exponential_jitter

from oaebu_workflows.airflow_pools import CrossrefEventsPool
from oaebu_workflows.config import schema_folder as default_schema_folder
from oaebu_workflows.config import sql_folder
from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from oaebu_workflows.onix_workflow.onix_work_aggregation import BookWorkAggregator, BookWorkFamilyAggregator
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.utils.dag_run_sensor import DagRunSensor
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import save_jsonl_gz
from observatory.platform.utils.url_utils import get_user_agent, retry_get_url
from observatory.platform.gcs import gcs_upload_files, gcs_blob_uri, gcs_blob_name_from_path
from observatory.platform.utils.jinja2_utils import render_template
from observatory.platform.api import make_observatory_api
from observatory.platform.bigquery import (
    bq_load_table,
    bq_table_id,
    bq_sharded_table_id,
    bq_create_dataset,
    bq_create_table_from_query,
    bq_run_query,
    bq_select_table_shard_dates,
    bq_create_view,
    bq_find_schema,
)
from observatory.platform.workflows.workflow import (
    SnapshotRelease,
    Workflow,
    make_snapshot_date,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


CROSSREF_EVENT_URL_TEMPLATE = (
    "https://api.eventdata.crossref.org/v1/events?mailto={mailto}"
    "&from-collected-date={start_date}&until-collected-date={end_date}&rows=1000"
    "&obj-id={doi}"
)


class OnixWorkflowRelease(SnapshotRelease):
    """
    Release information for OnixWorkflow.
    """

    def __init__(
        self,
        *,
        dag_id: str,
        run_id: str,
        snapshot_date: pendulum.DateTime,
        onix_snapshot_date: pendulum.DateTime,
        crossref_master_snapshot_date: pendulum.DateTime,
    ):
        """
        Construct the OnixWorkflow Release
        :param dag_id: DAG ID.
        :param release_date: The date of the partition/release
        :param onix_snapshot_date: The ONIX snapshot/release date.
        :param crossref_master_snapshot_date: The release date/suffix of the crossref master table

        """
        super().__init__(dag_id=dag_id, run_id=run_id, snapshot_date=snapshot_date)
        self.onix_snapshot_date = onix_snapshot_date
        self.crossref_master_snapshot_date = crossref_master_snapshot_date

        # Files
        self.workslookup_path = os.path.join(self.transform_folder, f"worksid.jsonl.gz")
        self.workslookup_errors_path = os.path.join(self.transform_folder, f"worksid_errors.jsonl.gz")
        self.worksfamilylookup_path = os.path.join(self.transform_folder, f"workfamilyid.jsonl.gz")
        self.crossref_metadata_path = os.path.join(self.transform_folder, f"crossref_metadata.jsonl.gz")
        self.crossref_events_path = os.path.join(self.transform_folder, f"crossref_events.jsonl.gz")


class OnixWorkflow(Workflow):
    """This workflow telescope:
    1. [Not implemented] Creates an ISBN13-> internal identifier lookup table.
    2. Creates an ISBN13 -> WorkID lookup table.
      a. Aggregates Product records into Work clusters.
      b. Writes the lookup table to BigQuery.
      c. Writes an error table to BigQuery.
    3. Create an ISBN13 -> Work Family ID lookup table.  Clusters editions together.
      a. Aggregate Works into Work Families.
      b. Writes the lookup table to BigQuery.
    4. Create OAEBU intermediate tables.
      a. For each data partner, create new tables in oaebu_intermediate dataset where existing tables are augmented
         with work_id and work_family_id columns.
    5. Create OAEBU QA tables for looking at metrics and problems arising from the data sets (and for eventual automatic reporting).
    """

    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        metadata_partner: Union[str, OaebuPartner],
        # Bigquery parameters
        bq_master_crossref_project_id: str = "academic-observatory",
        bq_master_crossref_dataset_id: str = "crossref_metadata",
        bq_oaebu_crossref_dataset_id: str = "crossref",
        bq_master_crossref_metadata_table_name: str = "crossref_metadata",
        bq_oaebu_crossref_metadata_table_name: str = "crossref_metadata",
        bq_crossref_events_table_name: str = "crossref_events",
        bq_country_project_id: str = "oaebu-public-data",
        bq_country_dataset_id: str = "oaebu_reference",
        bq_subject_project_id: str = "oaebu-public-data",
        bq_subject_dataset_id: str = "oaebu_reference",
        bq_book_table_name: str = "book",
        bq_book_product_table_name: str = "book_product",
        bq_oaebu_data_qa_dataset: str = "oaebu_data_qa",
        bq_oaebu_latest_data_qa_dataset: str = "oaebu_data_qa_latset",
        bq_onix_workflow_dataset: str = "onix_workflow",
        bq_oaebu_intermediate_dataset: str = "oaebu_intermediate",
        bq_oaebu_dataset: str = "oaebu",
        bq_oaebu_export_dataset: str = "data_export",
        bq_oaebu_latest_export_dataset: str = "data_export_latest",
        bq_worksid_table_name: str = "onix_workid_isbn",
        bq_worksid_error_table_name: str = "onix_workid_isbn_errors",
        bq_workfamilyid_table_name: str = "onix_workfamilyid_isbn",
        bq_dataset_description: str = "ONIX workflow tables",
        oaebu_intermediate_match_suffix: str = "_matched",
        # Run parameters
        data_partners: List[Union[str, OaebuPartner]] = None,
        ga3_views_field="page_views",
        schema_folder: str = default_schema_folder(workflow_module="onix_workflow"),
        mailto: str = "agent@observatory.academy",
        crossref_start_date: pendulum.DateTime = pendulum.datetime(2018, 5, 14),
        api_dataset_id: str = "onix_workflow",
        max_threads: int = 2 * os.cpu_count() - 1,
        # Ariflow parameters
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        sensor_dag_ids: List[str] = None,
        catchup: Optional[bool] = False,
        start_date: Optional[pendulum.DateTime] = pendulum.datetime(2022, 8, 1),
        schedule: Optional[str] = "@weekly",
    ):
        """
        Initialises the workflow object.

        :param dag_id: DAG ID.
        :param cloud_workspace: The CloudWorkspace object for this DAG

        :param bq_master_crossref_project_id: GCP project ID of crossref master data
        :param bq_master_crossref_dataset_id: GCP dataset ID of crossref master data
        :param bq_oaebu_crossref_dataset_id: GCP dataset ID of crossref OAeBU data
        :param bq_master_crossref_metadata_table_name: The name of the master crossref metadata table
        :param bq_oaebu_crossref_metadata_table_name: The name of the OAeBU crossref metadata table
        :param bq_crossref_events_table_name: The name of the crossref events table
        :param bq_country_project_id: GCP project ID of the country table
        :param bq_country_dataset_id: GCP dataset containing the country table
        :param bq_subject_project_id: GCP project ID of the subject tables
        :param bq_subject_dataset_id: GCP dataset ID of the subject tables
        :param bq_book_table_name: The name of the book table
        :param bq_book_product_table_name: The name of the book product table
        :param bq_oaebu_data_qa_dataset: OAEBU Data QA dataset.
        :param bq_oaebu_latest_data_qa_dataset: OAEBU Data QA dataset with the latest data views
        :param bq_onix_workflow_dataset: Onix workflow dataset.
        :param bq_oaebu_intermediate_dataset: OAEBU intermediate dataset.
        :param bq_oaebu_dataset: OAEBU dataset.
        :param bq_oaebu_export_dataset: OAEBU data export dataset.
        :param bq_oaebu_latest_export_dataset: OAEBU data export dataset with the latest data views
        :param bq_worksid_table_name: table ID of the worksid table
        :param bq_worksid_error_table_name: table ID of the worksid error table
        :param bq_workfamilyid_table_name: table ID of the workfamilyid table
        :param bq_dataset_description: Description to give to the workflow tables
        :param oaebu_intermediate_match_suffix: Suffix to append to intermediate tables

        :param data_partners: OAEBU data sources.
        :param ga3_views_field: The name of the GA3 views field - should be either 'page_views' or 'unique_views'
        :param schema_folder: the SQL schema path.
        :param mailto: email address used to identify the user when sending requests to an API.
        :param crossref_start_date: The starting date of crossref's API calls
        :param api_dataset_id: The ID to store the dataset release in the API
        :param max_threads: The maximum number of threads to use for parallel tasks.

        :param observatory_api_conn_id: The connection ID for the observatory API
        :param sensor_dag_ids: Dag IDs for dependent tasks
        :param catchup: Whether to catch up missed DAG runs.
        :param start_date: Start date of the DAG.
        :param schedule: Scheduled interval for running the DAG.
        """

        if not sensor_dag_ids:
            sensor_dag_ids = []

        if data_partners is None:
            data_partners = list()

        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.metadata_partner = partner_from_str(metadata_partner, metadata_partner=True)
        # Bigquery projects, datasets and tables
        self.bq_master_crossref_project_id = bq_master_crossref_project_id
        self.bq_master_crossref_dataset_id = bq_master_crossref_dataset_id
        self.bq_oaebu_crossref_dataset_id = bq_oaebu_crossref_dataset_id
        self.bq_master_crossref_metadata_table_name = bq_master_crossref_metadata_table_name
        self.bq_oaebu_crossref_metadata_table_name = bq_oaebu_crossref_metadata_table_name
        self.bq_crossref_events_table_name = bq_crossref_events_table_name
        self.bq_country_project_id = bq_country_project_id
        self.bq_country_dataset_id = bq_country_dataset_id
        self.bq_subject_project_id = bq_subject_project_id
        self.bq_subject_dataset_id = bq_subject_dataset_id
        self.bq_book_table_name = bq_book_table_name
        self.bq_book_product_table_name = bq_book_product_table_name
        self.bq_oaebu_data_qa_dataset = bq_oaebu_data_qa_dataset
        self.bq_oaebu_latest_data_qa_dataset = bq_oaebu_latest_data_qa_dataset
        self.bq_onix_workflow_dataset = bq_onix_workflow_dataset
        self.bq_oaebu_intermediate_dataset = bq_oaebu_intermediate_dataset
        self.bq_oaebu_dataset = bq_oaebu_dataset
        self.bq_oaebu_export_dataset = bq_oaebu_export_dataset
        self.bq_oaebu_latest_export_dataset = bq_oaebu_latest_export_dataset
        self.bq_worksid_table_name = bq_worksid_table_name
        self.bq_worksid_error_table_name = bq_worksid_error_table_name
        self.bq_workfamilyid_table_name = bq_workfamilyid_table_name
        self.bq_dataset_description = bq_dataset_description
        self.oaebu_intermediate_match_suffix = oaebu_intermediate_match_suffix
        # Run parameters
        self.data_partners = [partner_from_str(p) for p in data_partners]
        self.ga3_views_field = ga3_views_field
        self.schema_folder = schema_folder
        self.mailto = mailto
        self.crossref_start_date = crossref_start_date
        self.api_dataset_id = api_dataset_id
        self.max_threads = max_threads
        # Airflow Parameters
        self.observatory_api_conn_id = observatory_api_conn_id
        self.sensor_dag_ids = sensor_dag_ids
        self.catchup = catchup
        self.start_date = start_date
        self.schedule = schedule

        # Initialise Telesecope base class
        super().__init__(
            dag_id=self.dag_id,
            start_date=start_date,
            schedule=schedule,
            catchup=catchup,
            airflow_conns=[observatory_api_conn_id],
            tags=["oaebu"],
        )

        check_workflow_inputs(self)

        # Create pools for crossref API calls (if they don't exist)
        # Pools are necessary to throttle the maxiumum number of requests we can make per second and avoid 429 errors
        crossref_events_pool = CrossrefEventsPool(pool_slots=15)
        crossref_events_pool.create_pool()

        # Wait for external workflows to finish
        with self.parallel_tasks():
            for ext_dag_id in self.sensor_dag_ids:
                sensor = DagRunSensor(
                    task_id=f"{ext_dag_id}_sensor",
                    external_dag_id=ext_dag_id,
                    mode="reschedule",
                    duration=timedelta(days=7),  # Look back up to 7 days from execution date
                    poke_interval=int(timedelta(hours=1).total_seconds()),  # Check at this interval if dag run is ready
                    timeout=int(timedelta(days=2).total_seconds()),  # Sensor will fail after 2 days of waiting
                )
                self.add_operator(sensor)

        # Aggregate Works
        self.add_task(self.aggregate_works)
        self.add_task(self.upload_aggregation_tables)
        self.add_task(self.bq_load_aggregations)

        # Create crossref metadata and event tables
        self.add_task(self.create_oaebu_crossref_metadata_table)
        self.add_task(
            self.create_oaebu_crossref_events_table,
            pool=crossref_events_pool.pool_name,
            pool_slots=min(self.max_threads, crossref_events_pool.pool_slots),
        )

        # Create book table
        self.add_task(self.create_oaebu_book_table)

        # Create OAEBU Intermediate tables (not for onix partner types)
        with self.parallel_tasks():
            for data_partner in self.data_partners:
                task_id = f"create_oaebu_intermediate_table_{data_partner.bq_table_name}"

                self.add_task(
                    self.create_oaebu_intermediate_table,
                    op_kwargs=dict(
                        orig_project_id=self.cloud_workspace.project_id,
                        orig_dataset=data_partner.bq_dataset_id,
                        orig_table=data_partner.bq_table_name,
                        orig_isbn=data_partner.isbn_field_name,
                        sharded=data_partner.sharded,
                    ),
                    task_id=task_id,
                )

        # Create OAEBU tables
        self.add_task(self.create_oaebu_book_product_table)

        # Create QA metrics tables
        self.create_oaebu_data_qa_tasks()

        # Create OAEBU Elastic Export tables
        self.create_oaebu_export_tasks()

        # Create the (non-sharded) views of the sharded tables
        self.add_task(self.create_oaebu_latest_views)

        # Final tasks
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> OnixWorkflowRelease:
        """Creates a release object.
        :param kwargs: From Airflow. Contains the execution_date.
        :return: an OnixWorkflowRelease object.
        """

        # Make snapshot date
        snapshot_date = make_snapshot_date(**kwargs)

        # Get ONIX release date
        onix_table_id = bq_table_id(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.metadata_partner.bq_dataset_id,
            table_id=self.metadata_partner.bq_table_name,
        )
        onix_snapshot_dates = bq_select_table_shard_dates(table_id=onix_table_id, end_date=snapshot_date)
        assert len(onix_snapshot_dates), "OnixWorkflow.make_release: no ONIX releases found"
        onix_snapshot_date = onix_snapshot_dates[0]  # Get most recent snapshot

        # Get Crossref Metadata release date
        crossref_table_id = bq_table_id(
            project_id=self.bq_master_crossref_project_id,
            dataset_id=self.bq_master_crossref_dataset_id,
            table_id=self.bq_master_crossref_metadata_table_name,
        )
        crossref_metadata_snapshot_dates = bq_select_table_shard_dates(
            table_id=crossref_table_id, end_date=snapshot_date
        )
        assert len(crossref_metadata_snapshot_dates), "OnixWorkflow.make_release: no Crossref Metadata releases found"
        crossref_master_snapshot_date = crossref_metadata_snapshot_dates[0]  # Get most recent snapshot

        # Make the release object
        return OnixWorkflowRelease(
            dag_id=self.dag_id,
            run_id=kwargs["run_id"],
            snapshot_date=snapshot_date,
            onix_snapshot_date=onix_snapshot_date,
            crossref_master_snapshot_date=crossref_master_snapshot_date,
        )

    def aggregate_works(self, release: OnixWorkflowRelease, **kwargs):
        """Fetches the ONIX product records from our ONIX database, aggregates them into works, workfamilies,
        and outputs it into jsonl files."""

        # Fetch ONIX data
        sharded_onix_table = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.onix_snapshot_date,
        )
        products = get_onix_records(sharded_onix_table)

        # Aggregate into works
        agg = BookWorkAggregator(products)
        works = agg.aggregate()
        lookup_table = agg.get_works_lookup_table()
        save_jsonl_gz(release.workslookup_path, lookup_table)

        # Save errors from aggregation
        error_table = [{"Error": error} for error in agg.errors]
        save_jsonl_gz(release.workslookup_errors_path, error_table)

        # Aggregate work families
        agg = BookWorkFamilyAggregator(works)
        agg.aggregate()
        lookup_table = agg.get_works_family_lookup_table()
        save_jsonl_gz(release.worksfamilylookup_path, lookup_table)

    def upload_aggregation_tables(self, release: OnixWorkflowRelease, **kwargs):
        """Upload the aggregation tables and error tables to a GCP bucket in preparation for BQ loading."""
        files = [release.workslookup_path, release.workslookup_errors_path, release.worksfamilylookup_path]
        gcs_upload_files(bucket_name=self.cloud_workspace.transform_bucket, file_paths=files)

    def bq_load_aggregations(self, release: OnixWorkflowRelease, **kwargs):
        """Loads the 'WorkID lookup', 'WorkID lookup table errors' and 'WorkFamilyID lookup' tables into BigQuery."""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_onix_workflow_dataset,
            location=self.cloud_workspace.data_location,
            description="Onix Workflow Aggregations",
        )

        aggregation_paths = [release.workslookup_path, release.workslookup_errors_path, release.worksfamilylookup_path]
        aggregation_tables = [
            self.bq_worksid_table_name,
            self.bq_worksid_error_table_name,
            self.bq_workfamilyid_table_name,
        ]
        for path, table_name in zip(aggregation_paths, aggregation_tables):
            uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(path))
            table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id, self.bq_onix_workflow_dataset, table_name, release.snapshot_date
            )
            state = bq_load_table(
                uri=uri,
                table_id=table_id,
                schema_file_path=bq_find_schema(path=self.schema_folder, table_name=table_name),
                source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
            )
            set_task_state(state, kwargs["ti"].task_id, release=release)

    def create_oaebu_crossref_metadata_table(self, release: OnixWorkflowRelease, **kwargs):
        """Creates the crossref metadata table by querying the AO master table and matching on this publisher's ISBNs

        :param release: The onix workflow release object
        """
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_crossref_dataset_id,
            location=self.cloud_workspace.data_location,
            description="Data from Crossref sources",
        )

        onix_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.onix_snapshot_date,
        )
        master_crossref_metadata_table_id = bq_sharded_table_id(
            self.bq_master_crossref_project_id,
            self.bq_master_crossref_dataset_id,
            self.bq_master_crossref_metadata_table_name,
            release.crossref_master_snapshot_date,
        )
        sql = render_template(
            os.path.join(sql_folder(workflow_module="onix_workflow"), "crossref_metadata_filter_isbn.sql.jinja2"),
            onix_table_id=onix_table_id,
            crossref_metadata_table_id=master_crossref_metadata_table_id,
        )
        print("Creating crossref metadata table from master table")
        schema_file_path = bq_find_schema(
            path=self.schema_folder, table_name=self.bq_oaebu_crossref_metadata_table_name
        )
        oaebu_crossref_metadata_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_crossref_dataset_id,
            self.bq_oaebu_crossref_metadata_table_name,
            release.snapshot_date,
        )
        state = bq_create_table_from_query(
            sql=sql, table_id=oaebu_crossref_metadata_table_id, schema_file_path=schema_file_path
        )
        set_task_state(state, kwargs["ti"].task_id, release=release)

    def create_oaebu_crossref_events_table(self, release: OnixWorkflowRelease, **kwargs):
        """Download, transform, upload and create a table for crossref events"""

        # Get the unique dois from the metadata table
        metadata_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_crossref_dataset_id,
            self.bq_oaebu_crossref_metadata_table_name,
            release.snapshot_date,
        )
        dois = dois_from_table(metadata_table_id, doi_column_name="DOI", distinct=True)

        # Download and transform all events
        start_date = self.crossref_start_date
        end_date = release.snapshot_date.subtract(days=1).date()
        events = download_crossref_events(dois, start_date, end_date, self.mailto, max_threads=self.max_threads)
        events = transform_crossref_events(events, max_threads=self.max_threads)

        # Zip and upload to google cloud
        save_jsonl_gz(release.crossref_events_path, events)
        gcs_upload_files(bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.crossref_events_path])
        uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.crossref_events_path))
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_crossref_dataset_id,
            self.bq_crossref_events_table_name,
            release.snapshot_date,
        )
        state = bq_load_table(
            uri=uri,
            table_id=table_id,
            schema_file_path=bq_find_schema(path=self.schema_folder, table_name=self.bq_crossref_events_table_name),
            source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
        )
        set_task_state(state, kwargs["ti"].task_id, release=release)

    def create_oaebu_book_table(self, release: OnixWorkflowRelease, **kwargs):
        """Create the oaebu book table using the crossref event and metadata tables"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Tables",
        )
        book_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id, self.bq_oaebu_dataset, self.bq_book_table_name, release.snapshot_date
        )
        crossref_metadata_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_crossref_dataset_id,
            self.bq_oaebu_crossref_metadata_table_name,
            release.snapshot_date,
        )
        crossref_events_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_crossref_dataset_id,
            self.bq_crossref_events_table_name,
            release.snapshot_date,
        )
        sql = render_template(
            os.path.join(sql_folder(workflow_module="onix_workflow"), "create_book.sql.jinja2"),
            crossref_events_table_id=crossref_events_table_id,
            crossref_metadata_table_id=crossref_metadata_table_id,
        )
        status = bq_create_table_from_query(
            sql=sql,
            table_id=book_table_id,
            schema_file_path=bq_find_schema(path=self.schema_folder, table_name=self.bq_book_table_name),
        )
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def create_oaebu_intermediate_table(
        self,
        release: OnixWorkflowRelease,
        *,
        orig_project_id: str,
        orig_dataset: str,
        orig_table: str,
        orig_isbn: str,
        sharded: bool,
        **kwargs,
    ):
        """Create an intermediate oaebu table.  They are of the form datasource_matched<date>
        :param release: Onix workflow release information.
        :param orig_project_id: Project ID for the partner data.
        :param orig_dataset: Dataset ID for the partner data.
        :param orig_table: Table ID for the partner data.
        :param orig_isbn: Name of the ISBN field in the partner data table.
        :param sharded: Whether the data partner table is sharded
        """
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_intermediate_dataset,
            location=self.cloud_workspace.data_location,
            description="Intermediate OAEBU Tables",
        )
        orig_table_id = (
            bq_sharded_table_id(orig_project_id, orig_dataset, orig_table, release.snapshot_date)
            if sharded
            else bq_table_id(orig_project_id, orig_dataset, orig_table)
        )
        output_table_name = f"{orig_table}{self.oaebu_intermediate_match_suffix}"
        template_path = os.path.join(
            sql_folder(workflow_module="onix_workflow"), "assign_workid_workfamilyid.sql.jinja2"
        )
        output_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_intermediate_dataset,
            output_table_name,
            release.snapshot_date,
        )
        wid_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_onix_workflow_dataset,
            self.bq_worksid_table_name,
            release.snapshot_date,
        )
        wfam_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_onix_workflow_dataset,
            self.bq_workfamilyid_table_name,
            release.snapshot_date,
        )

        # Make the table from SQL query
        sql = render_template(
            template_path,
            orig_table_id=orig_table_id,
            orig_isbn=orig_isbn,
            wid_table_id=wid_table_id,
            wfam_table_id=wfam_table_id,
        )
        status = bq_create_table_from_query(sql=sql, table_id=output_table_id)
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def create_oaebu_book_product_table(
        self,
        release: OnixWorkflowRelease,
        **kwargs,
    ):
        """Create the Book Product Table"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Tables",
        )
        data_partner_datasets = {data.type_id: data.bq_dataset_id for data in self.data_partners}
        google_analytics_dataset = data_partner_datasets.get("google_analytics3")
        google_books_dataset = data_partner_datasets.get("google_books_traffic")
        jstor_dataset = data_partner_datasets.get("jstor_country")
        irus_oapen_dataset = data_partner_datasets.get("irus_oapen")
        irus_fulcrum_dataset = data_partner_datasets.get("irus_fulcrum")
        ucl_discovery_dataset = data_partner_datasets.get("ucl_discovery")
        internet_archive_dataset = data_partner_datasets.get("internet_archive")
        worldreader_dataset = data_partner_datasets.get("worldreader")

        # Create matched tables for supplied data partners
        google_analytics3_table_id = "empty_google_analytics3"
        google_books_sales_table_id = "empty_google_books_sales"
        google_books_traffic_table_id = "empty_google_books_traffic"
        jstor_country_table_id = "empty_jstor_country"
        jstor_institution_table_id = "empty_jstor_institution"
        irus_oapen_table_id = "empty_irus_oapen"
        irus_fulcrum_table_id = "empty_irus_fulcrum"
        ucl_discovery_table_id = "empty_ucl_discovery"
        internet_archive_table_id = "empty_internet_archive"
        worldreader_table_id = "empty_worldreader"
        if google_analytics_dataset:
            google_analytics3_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "google_analytics3_matched",
                release.snapshot_date,
            )
        if google_books_dataset:
            google_books_sales_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "google_books_sales_matched",
                release.snapshot_date,
            )
            google_books_traffic_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "google_books_traffic_matched",
                release.snapshot_date,
            )
        if jstor_dataset:
            jstor_country_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "jstor_country_matched",
                release.snapshot_date,
            )
            jstor_institution_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "jstor_institution_matched",
                release.snapshot_date,
            )
        if irus_oapen_dataset:
            irus_oapen_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "irus_oapen_matched",
                release.snapshot_date,
            )
        if ucl_discovery_dataset:
            ucl_discovery_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "ucl_discovery_matched",
                release.snapshot_date,
            )
        if irus_fulcrum_dataset:
            irus_fulcrum_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "irus_fulcrum_matched",
                release.snapshot_date,
            )
        if internet_archive_dataset:
            internet_archive_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "internet_archive_matched",
                release.snapshot_date,
            )
        if worldreader_dataset:
            worldreader_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_intermediate_dataset,
                "worldreader_matched",
                release.snapshot_date,
            )
        workid_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_onix_workflow_dataset,
            self.bq_worksid_table_name,
            release.snapshot_date,
        )
        workfamilyid_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_onix_workflow_dataset,
            self.bq_workfamilyid_table_name,
            release.snapshot_date,
        )
        book_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id, self.bq_oaebu_dataset, self.bq_book_table_name, release.snapshot_date
        )
        onix_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.onix_snapshot_date,
        )
        country_table_id = bq_table_id(self.bq_country_project_id, self.bq_country_dataset_id, "country")
        template_path = os.path.join(sql_folder(workflow_module="onix_workflow"), "create_book_products.sql.jinja2")
        sql = render_template(
            template_path,
            onix_table_id=onix_table_id,
            google_analytics3_table_id=google_analytics3_table_id,
            google_books_sales_table_id=google_books_sales_table_id,
            google_books_traffic_table_id=google_books_traffic_table_id,
            jstor_country_table_id=jstor_country_table_id,
            jstor_institution_table_id=jstor_institution_table_id,
            irus_oapen_table_id=irus_oapen_table_id,
            irus_fulcrum_table_id=irus_fulcrum_table_id,
            ucl_discovery_table_id=ucl_discovery_table_id,
            internet_archive_table_id=internet_archive_table_id,
            worldreader_table_id=worldreader_table_id,
            book_table_id=book_table_id,
            country_table_id=country_table_id,
            workid_table_id=workid_table_id,
            workfamilyid_table_id=workfamilyid_table_id,
            ga3_views_field=self.ga3_views_field,
            onix_workflow=True,
        )
        schema_file_path = bq_find_schema(path=self.schema_folder, table_name=self.bq_book_product_table_name)
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_dataset,
            self.bq_book_product_table_name,
            release.snapshot_date,
        )
        status = bq_create_table_from_query(sql=sql, table_id=table_id, schema_file_path=schema_file_path)
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def export_oaebu_table(
        self,
        release: OnixWorkflowRelease,
        **kwargs,
    ):
        """Create an intermediate oaebu table. They are of the form datasource_matched<date>"""
        output_table: str = kwargs["output_table"]
        query_template: str = kwargs["query_template"]
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_export_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Tables for Dashboarding",
        )
        output_table_name = f"{self.cloud_workspace.project_id.replace('-', '_')}_{output_table}"
        output_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id, self.bq_oaebu_export_dataset, output_table_name, release.snapshot_date
        )
        template_path = os.path.join(sql_folder(workflow_module="onix_workflow"), query_template)
        schema_file_path = bq_find_schema(
            path=self.schema_folder,
            table_name=output_table,
            prefix="oaebu_publisher_",
        )
        book_product_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_dataset,
            self.bq_book_product_table_name,
            release.snapshot_date,
        )
        country_table_id = bq_table_id(self.bq_country_project_id, self.bq_country_dataset_id, "country")
        bic_table_id = bq_table_id(self.bq_subject_project_id, self.bq_subject_dataset_id, "bic_lookup")
        bisac_table_id = bq_table_id(self.bq_subject_project_id, self.bq_subject_dataset_id, "bisac_lookup")
        thema_table_id = bq_table_id(self.bq_subject_project_id, self.bq_subject_dataset_id, "thema_lookup")
        sql = render_template(
            template_path,
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_dataset,
            release=release.snapshot_date,
            book_product_table_id=book_product_table_id,
            country_table_id=country_table_id,
            bic_table_id=bic_table_id,
            bisac_table_id=bisac_table_id,
            thema_table_id=thema_table_id,
        )
        status = bq_create_table_from_query(sql=sql, table_id=output_table_id, schema_file_path=schema_file_path)
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def export_oaebu_qa_metrics(
        self,
        release: OnixWorkflowRelease,
        **kwargs,
    ):
        """Create the unmatched metrics table"""
        data_partner_isbns = {data.type_id: data.isbn_field_name for data in self.data_partners}
        data_partner_tables = {data.type_id: data.bq_table_name for data in self.data_partners}
        google_analytics3_table = data_partner_tables.get("google_analytics3")
        google_books_table = data_partner_tables.get("google_books_traffic")
        jstor_table = data_partner_tables.get("jstor_country")
        irus_oapen_table = data_partner_tables.get("irus_oapen")
        ucl_table = data_partner_tables.get("ucl_discovery")
        fulcrum_table = data_partner_tables.get("irus_fulcrum")

        google_analytics3_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{google_analytics3_table}_unmatched_{data_partner_isbns['google_analytics3']}",
                release.snapshot_date,
            )
            if google_analytics3_table
            else None
        )
        google_books_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{google_books_table}_unmatched_{data_partner_isbns['google_books_traffic']}",
                release.snapshot_date,
            )
            if google_books_table
            else None
        )
        jstor_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{jstor_table}_unmatched_{data_partner_isbns['jstor_country']}",
                release.snapshot_date,
            )
            if jstor_table
            else None
        )
        irus_oapen_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{irus_oapen_table}_unmatched_{data_partner_isbns['irus_oapen']}",
                release.snapshot_date,
            )
            if irus_oapen_table
            else None
        )
        ucl_discovery_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{ucl_table}_unmatched_{data_partner_isbns['ucl_discovery']}",
                release.snapshot_date,
            )
            if ucl_table
            else None
        )
        fulcrum_unmatched_table_id = (
            bq_sharded_table_id(
                self.cloud_workspace.project_id,
                self.bq_oaebu_data_qa_dataset,
                f"{fulcrum_table}_unmatched_{data_partner_isbns['irus_fulcrum']}",
                release.snapshot_date,
            )
            if fulcrum_table
            else None
        )
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_export_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Tables for Dashboarding",
        )
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_export_dataset,
            f"{self.cloud_workspace.project_id.replace('-', '_')}_unmatched_book_metrics",
            release.snapshot_date,
        )
        template_path = os.path.join(sql_folder(workflow_module="onix_workflow"), "export_unmatched_metrics.sql.jinja2")
        sql = render_template(
            template_path,
            google_analytics3_unmatched_table_id=google_analytics3_unmatched_table_id,
            google_books_unmatched_table_id=google_books_unmatched_table_id,
            jstor_unmatched_table_id=jstor_unmatched_table_id,
            irus_oapen_unmatched_table_id=irus_oapen_unmatched_table_id,
            ucl_discovery_unmatched_table_id=ucl_discovery_unmatched_table_id,
            fulcrum_unmatched_table_id=fulcrum_unmatched_table_id,
            google_analytics3_isbn=data_partner_isbns.get("google_analytics3"),
            google_books_isbn=data_partner_isbns.get("google_books_traffic"),
            jstor_isbn=data_partner_isbns.get("jstor_country"),
            irus_oapen_isbn=data_partner_isbns.get("irus_oapen"),
            ucl_discovery_isbn=data_partner_isbns.get("ucl_discovery"),
            fulcrum_isbn=data_partner_isbns.get("irus_fulcrum"),
        )
        status = bq_create_table_from_query(sql=sql, table_id=table_id)
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def create_oaebu_export_tasks(self):
        """Create tasks for exporting final metrics from our OAEBU data.  It will create output tables in the oaebu_elastic dataset."""
        export_tables = [
            {"output_table": "book_product_list", "query_template": "export_book_list.sql.jinja2"},
            {
                "output_table": "book_product_metrics",
                "query_template": "export_book_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_metrics_country",
                "query_template": "export_book_metrics_country.sql.jinja2",
            },
            {
                "output_table": "book_product_metrics_institution",
                "query_template": "export_book_metrics_institution.sql.jinja2",
            },
            {
                "output_table": "institution_list",
                "query_template": "export_institution_list.sql.jinja2",
            },
            {
                "output_table": "book_product_metrics_city",
                "query_template": "export_book_metrics_city.sql.jinja2",
            },
            {
                "output_table": "book_product_metrics_events",
                "query_template": "export_book_metrics_event.sql.jinja2",
            },
            {
                "output_table": "book_product_publisher_metrics",
                "query_template": "export_book_publisher_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_subject_bic_metrics",
                "query_template": "export_book_subject_bic_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_subject_bisac_metrics",
                "query_template": "export_book_subject_bisac_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_subject_thema_metrics",
                "query_template": "export_book_subject_thema_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_year_metrics",
                "query_template": "export_book_year_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_subject_year_metrics",
                "query_template": "export_book_subject_year_metrics.sql.jinja2",
            },
            {
                "output_table": "book_product_author_metrics",
                "query_template": "export_book_author_metrics.sql.jinja2",
            },
        ]

        # Create each export table in BiqQuery
        with self.parallel_tasks():
            for export_table in export_tables:
                task_id = f"export_oaebu_table_{export_table['output_table']}"
                self.add_task(
                    self.export_oaebu_table,
                    op_kwargs=dict(
                        output_table=export_table["output_table"], query_template=export_table["query_template"]
                    ),
                    task_id=task_id,
                )

        # Export QA Metrics
        self.add_task(self.export_oaebu_qa_metrics)

    def create_oaebu_data_qa_tasks(self):
        """Create tasks for outputing QA metrics from our OAEBU data.
        It will create output tables in the oaebu_data_qa dataset.
        """
        with self.parallel_tasks():
            self.add_task(self.create_oaebu_data_qa_isbn_onix)
            self.add_task(self.create_oaebu_data_qa_onix_aggregate)
            for data_partner in self.data_partners:
                task_id_invalid_isbn = f"create_oaebu_data_qa_isbn_{data_partner.bq_table_name}"
                task_id_unmatched_isbn = f"create_oaebu_data_qa_intermediate_unmatched_{data_partner.bq_table_name}"
                self.add_task(
                    self.create_oaebu_data_qa_intermediate_unmatched_workid,
                    op_kwargs=dict(data_partner=data_partner),
                    task_id=task_id_unmatched_isbn,
                )
                # JSTOR has QA for EISBN and collates country + institution
                if data_partner.type_id.startswith("jstor"):
                    self.add_task(
                        self.create_oaebu_data_qa_isbn,
                        op_kwargs=dict(
                            data_partner=data_partner,
                            table_name="jstor_invalid_isbn",
                            isbn=data_partner.isbn_field_name,
                        ),
                        task_id=task_id_invalid_isbn,
                    )
                    self.add_task(
                        self.create_oaebu_data_qa_isbn,
                        op_kwargs=dict(
                            data_partner=data_partner,
                            table_name="jstor_invalid_eisbn",
                            isbn="eISBN",
                        ),
                        task_id=task_id_invalid_isbn.replace("isbn", "eisbn"),
                    )
                else:
                    self.add_task(
                        self.create_oaebu_data_qa_isbn,
                        op_kwargs=dict(
                            data_partner=data_partner,
                            table_name=f"{data_partner.type_id}_invalid_isbn",
                            isbn=data_partner.isbn_field_name,
                        ),
                        task_id=task_id_invalid_isbn,
                    )

    def create_oaebu_latest_views(self, release: OnixWorkflowRelease, **kwargs):
        """Create views of the latest data export tables in bigquery"""
        create_latest_views_from_dataset(
            project_id=self.cloud_workspace.project_id,
            from_dataset=self.bq_oaebu_data_qa_dataset,
            to_dataset=self.bq_oaebu_latest_data_qa_dataset,
            date_match=release.snapshot_date.strftime("%Y%m%d"),
            data_location=self.cloud_workspace.data_location,
        )
        create_latest_views_from_dataset(
            project_id=self.cloud_workspace.project_id,
            from_dataset=self.bq_oaebu_export_dataset,
            to_dataset=self.bq_oaebu_latest_export_dataset,
            date_match=release.snapshot_date.strftime("%Y%m%d"),
            data_location=self.cloud_workspace.data_location,
        )

    def create_oaebu_data_qa_onix_aggregate(self, release: OnixWorkflowRelease, **kwargs):
        """Create a bq table of some aggregate metrics for the ONIX data set."""
        onix_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.onix_snapshot_date,
        )
        output_table_name = "onix_aggregate_metrics"
        table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_data_qa_dataset,
            output_table_name,
            release.snapshot_date,
        )
        template_path = os.path.join(sql_folder(workflow_module="onix_workflow"), "onix_aggregate_metrics.sql.jinja2")
        sql = render_template(template_path, table_id=onix_table_id, isbn="ISBN13")
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_data_qa_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Quality Analysis Tables",
        )
        schema_file_path = bq_find_schema(path=self.schema_folder, table_name=output_table_name)
        status = bq_create_table_from_query(
            sql=sql,
            table_id=table_id,
            schema_file_path=schema_file_path,
        )
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def create_oaebu_data_qa_isbn_onix(self, release: OnixWorkflowRelease, **kwargs):
        """Create a BQ table of invalid ISBNs for the ONIX feed that can be fed back to publishers.
        No attempt is made to normalise the string so we catch as many string issues as we can.
        """
        orig_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.metadata_partner.bq_dataset_id,
            self.metadata_partner.bq_table_name,
            release.onix_snapshot_date,
        )
        output_table_name = "onix_invalid_isbn"
        output_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id, self.bq_oaebu_data_qa_dataset, output_table_name, release.snapshot_date
        )
        status = self.oaebu_data_qa_validate_isbn(
            orig_table_id=orig_table_id,
            output_table_id=output_table_id,
            isbn="ISBN13",
            schema_file_path=bq_find_schema(path=self.schema_folder, table_name=output_table_name),
        )
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def oaebu_data_qa_validate_isbn(
        self,
        *,
        orig_table_id: str,
        output_table_id: str,
        isbn: str,
        schema_file_path: str = None,
    ) -> bool:
        """Create a BQ table of invalid ISBNs for the ONIX feed that can be fed back to publishers.
        No attempt is made to normalise the string so we catch as many string issues as we can.

        :param orig_table_id: Fully qualified table ID of the source data
        :param output_table_id: Fully qualified table ID for the output data.
        :param isbn: Name of the isbn field in source table.
        :param schema_file_path: The path of the schema file to use for the BigQuery upload
        :return: The status of the table creation
        """
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_data_qa_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Quality Analysis Tables",
        )
        sql_template = os.path.join(sql_folder(workflow_module="onix_workflow"), "validate_isbn.sql.jinja2")
        sql = get_isbn_utils_sql_string() + render_template(sql_template, table_id=orig_table_id, isbn=isbn)
        status = bq_create_table_from_query(sql=sql, table_id=output_table_id, schema_file_path=schema_file_path)
        return status

    def create_oaebu_data_qa_isbn(
        self,
        release: OnixWorkflowRelease,
        data_partner: OaebuPartner,
        table_name: str,
        isbn: str,
        **kwargs,
    ):
        """Create a BQ table of invalid ISBNs for the Google Analytics feed.
        No attempt is made to normalise the string so we catch as many string issues as we can.

        :param release: workflow release object.
        :param data_partner: OaebuPartner,
        :param table_name: The name of the table to create
        :param isbn: Name of the isbn field in source table.
        """

        if data_partner.sharded:
            orig_table_id = bq_sharded_table_id(
                self.cloud_workspace.project_id,
                data_partner.bq_dataset_id,
                data_partner.bq_table_name,
                release.snapshot_date,
            )
        else:
            orig_table_id = bq_table_id(
                self.cloud_workspace.project_id,
                data_partner.bq_dataset_id,
                data_partner.bq_table_name,
            )

        output_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_data_qa_dataset,
            table_name,
            release.snapshot_date,
        )
        # Validate the ISBN field
        status = self.oaebu_data_qa_validate_isbn(
            orig_table_id=orig_table_id,
            output_table_id=output_table_id,
            isbn=isbn,
        )
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def create_oaebu_data_qa_intermediate_unmatched_workid(
        self,
        release: OnixWorkflowRelease,
        data_partner: OaebuPartner,
        *args,
        **kwargs,
    ):
        """Create quality assurance metrics for the OAEBU intermediate tables.
        :param data_partner: The OaebuPartner to use
        """
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.bq_oaebu_data_qa_dataset,
            location=self.cloud_workspace.data_location,
            description="OAEBU Quality Analysis Tables",
        )
        template_path = os.path.join(
            sql_folder(workflow_module="onix_workflow"), "oaebu_intermediate_metrics.sql.jinja2"
        )
        intermediate_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id,
            self.bq_oaebu_intermediate_dataset,
            f"{data_partner.bq_table_name}_matched",
            release.snapshot_date,
        )
        sql = render_template(
            template_path,
            table_id=intermediate_table_id,
            isbn=data_partner.isbn_field_name,
            title=data_partner.title_field_name,
        )
        output_table_name = f"{data_partner.bq_table_name}_unmatched_{data_partner.isbn_field_name}"
        output_table_id = bq_sharded_table_id(
            self.cloud_workspace.project_id, self.bq_oaebu_data_qa_dataset, output_table_name, release.snapshot_date
        )
        status = bq_create_table_from_query(sql=sql, table_id=output_table_id)
        set_task_state(status, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, release: OnixWorkflowRelease, **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        dataset_release = DatasetRelease(
            dag_id=self.dag_id,
            dataset_id=self.api_dataset_id,
            dag_run_id=release.run_id,
            snapshot_date=release.snapshot_date,
            data_interval_start=kwargs["data_interval_start"],
            data_interval_end=kwargs["data_interval_end"],
        )
        api.post_dataset_release(dataset_release)

    def cleanup(self, release: OnixWorkflowRelease, **kwargs):
        """Cleanup temporary files."""
        cleanup(dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder)


def dois_from_table(table_id: str, doi_column_name: str = "DOI", distinct: str = True) -> List[str]:
    """
    Queries a metadata table to retrieve the unique DOIs. Provided the DOIs are not in a nested structure.

    :param metadata_table_id: The fully qualified ID of the metadata table on GCP
    :param doi_field_name: The name of the DOI column
    :param distinct: Whether to retrieve only unique DOIs
    :return: All DOIs present in the metadata table
    """
    select_field = f"DISTINCT({doi_column_name})" if distinct else doi_column_name
    sql = f"SELECT {select_field} FROM `{table_id}`"
    query_results = bq_run_query(sql)
    dois = [r["DOI"] for r in query_results]
    return dois


def download_crossref_events(
    dois: List[str],
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime,
    mailto: str,
    max_threads: int = 1,
) -> List[dict]:
    """
    Spawns multiple threads to download event data (DOI and publisher only) for each doi supplied.
    The url template was made with reference to the crossref event api:
    https://www.eventdata.crossref.org/guide/service/query-api/
    Note that the max_threads will cap at 15 because the events API will return a 429 if more than 15 requests are made
    per second. Each API request happens to take roughly 1 second. Having more threadsthan necessary slows down the
    download process as the retry script will wait a minimum of two seconds between each attempt.

    :param dois: The list of DOIs to download the events for
    :param start_date: The start date for events we're interested in
    :param end_date: The end date for events we're interested in
    :param mailto: The email to use as a reference for who is requesting the data
    :param max_threads: The maximum threads to spawn for the downloads.
    :return: All events for the input DOIs
    """
    event_url_template = CROSSREF_EVENT_URL_TEMPLATE
    url_start_date = start_date.strftime("%Y-%m-%d")
    url_end_date = end_date.strftime("%Y-%m-%d")
    max_threads = min(max_threads, 15)

    event_urls = [
        event_url_template.format(doi=doi, mailto=mailto, start_date=url_start_date, end_date=url_end_date)
        for doi in dois
    ]

    print(f"Beginning crossref event data download from {len(event_urls)} URLs with {max_threads} workers")
    print(
        f"Downloading DOI data using URL: {event_url_template.format(doi='***', mailto=mailto, start_date=url_start_date, end_date=url_end_date)}"
    )
    all_events = []
    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        futures = []
        for i, url in enumerate(event_urls):
            futures.append(executor.submit(download_crossref_event_url, url, i=i))
        for future in as_completed(futures):
            all_events.extend(future.result())

    return all_events


def download_crossref_event_url(url: str, i: int = 0) -> List[dict]:
    """
    Downloads all crossref events from a url, iterating through pages if there is more than one

    :param url: The url send the request to
    :param i: Worker number
    :return: The events from this URL
    """
    events = []
    headers = {"User-Agent": get_user_agent(package_name="oaebu_workflows")}
    next_cursor, page_counts, total_events, page_events = download_crossref_page_events(url, headers)
    events.extend(page_events)
    total_counts = page_counts
    while next_cursor:
        tmp_url = url + f"&cursor={next_cursor}"
        next_cursor, page_counts, _, page_events = download_crossref_page_events(tmp_url, headers)
        total_counts += page_counts
        events.extend(page_events)
    print(f"{i + 1}: {url} successful")
    print(f"{i + 1}: Total no. events: {total_events}, downloaded " f"events: {total_counts}")
    return events


def download_crossref_page_events(url: str, headers: dict) -> Tuple[str, int, int, List[dict]]:
    """
    Download crossref events from a single page

    :param url: The url to send the request to
    :param headers: Headers to send with the request
    :return: The cursor, event counter, total number of events and the events for the URL
    """
    crossref_events_limiter()
    response = retry_get_url(url, num_retries=5, wait=wait_exponential_jitter(initial=0.5, max=60), headers=headers)
    response_json = response.json()
    total_events = response_json["message"]["total-results"]
    events = response_json["message"]["events"]
    next_cursor = response_json["message"]["next-cursor"]
    counter = len(events)

    return next_cursor, counter, total_events, events


@sleep_and_retry
@limits(calls=15, period=1)
def crossref_events_limiter():
    """ "Task to throttle the calls to the crossref events API"""
    return


def transform_crossref_events(events: List[dict], max_threads: int = 1) -> List[dict]:
    """
    Spawns workers to transforms crossref events

    :param all_events: A list of the events to transform
    :param max_threads: The maximum number of threads to utilise for the transforming process
    :return: transformed events, the order of the events in the input list is not preserved
    """
    print(f"Beginning crossref event transform with {max_threads} workers")
    transformed_events = []
    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        futures = []
        for event in events:
            futures.append(executor.submit(transform_event, event))
        for future in as_completed(futures):
            transformed_events.append(future.result())
    print("Crossref event transformation complete")
    return transformed_events


def transform_event(event: dict) -> dict:
    """Transform the dictionary with event data by replacing '-' with '_' in key names, converting all int values to
    string except for the 'total' field and parsing datetime columns for a valid datetime.

    :param event: The event dictionary
    :return: The transformed event dictionary
    """
    if isinstance(event, (str, int, float)):
        return event
    if isinstance(event, dict):
        new = event.__class__()
        for k, v in event.items():
            if isinstance(v, int) and k != "total":
                v = str(v)
            if k in ["timestamp", "occurred_at", "issued", "dateModified", "updated_date"]:
                try:
                    v = str(pendulum.parse(v))
                except ValueError:
                    v = "0001-01-01T00:00:00Z"

            # Replace hyphens with underscores for BigQuery compatibility
            k = k.replace("-", "_")

            # Replace @ symbol in keys left by DataCite between the 15 and 22 March 2019
            k = k.replace("@", "")

            new[k] = transform_event(v)
        return new


def create_latest_views_from_dataset(
    project_id: str, from_dataset: str, to_dataset: str, date_match: str, data_location: str, description: str = None
) -> None:
    """Creates views from all sharded tables from a dataset with a matching a date string.

    :param project_id: The project id
    :param from_dataset: The dataset containing the sharded tables
    :param to_dataset: The dataset to contain the views
    :param date_match: The date string to match. e.g. for a table named 'this_table20220101', this would be '20220101'
    :param data_location: The regional location of the data in google cloud
    :param description: The description for the views dataset
    """
    if description is None:
        description = "OAEBU Export Views for Dashboarding"
    # Make to_dataset if it doesn't exist
    bq_create_dataset(
        project_id=project_id,
        dataset_id=to_dataset,
        location=data_location,
        description=description,
    )

    # Get the tables from the from_dataset
    client = Client(project_id)
    tables = [t.table_id for t in client.list_tables(from_dataset)]

    # Find the tables with specified date string
    regex_string = rf"^\w+{date_match}\b"
    matched_tables = [re.findall(regex_string, t) for t in tables]
    matched_tables = [t[0] for t in matched_tables if t]
    assert len(matched_tables), f"No tables matching date {date_match} in dataset {project_id}.{from_dataset}"

    # Create all of the views
    for table in matched_tables:
        query = f"SELECT * FROM `{project_id}.{from_dataset}.{table}`"
        view_name = re.match(r"(.*?)\d{8}$", table).group(1)  # Drop the date from the table for the view name
        view_id = bq_table_id(project_id, to_dataset, view_name)
        bq_create_view(view_id=view_id, query=query)


def get_onix_records(table_id: str) -> List[dict]:
    """Fetch the latest onix snapshot from BigQuery.
    :param table_id: Fully qualified table ID.
    :return: List of onix product records.
    """

    sql = f"SELECT * FROM {table_id}"
    records = bq_run_query(sql)
    products = [{key: records[i][key] for key in records[i].keys()} for i in range(len(records))]
    return products


def get_isbn_utils_sql_string() -> str:
    """Load the ISBN utils sql functions.
    :return BQ SQL string.
    """

    isbn_utils_file = "isbn_utils.sql"
    isbn_utils_path = os.path.join(sql_folder(workflow_module="onix_workflow"), isbn_utils_file)
    with open(isbn_utils_path, "r") as f:
        isbn_utils_sql = f.read()

    return isbn_utils_sql



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien, Keegan Smith

import os
from datetime import timedelta
from unittest.mock import MagicMock, patch
import vcr
import shutil
from typing import List

import pendulum
from airflow.models import DagBag
from airflow.utils.state import State

from oaebu_workflows.config import schema_folder as default_schema_folder
from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from oaebu_workflows.onix_workflow.onix_workflow import (
    OnixWorkflow,
    OnixWorkflowRelease,
    CROSSREF_EVENT_URL_TEMPLATE,
    download_crossref_events,
    transform_crossref_events,
    transform_event,
    dois_from_table,
    download_crossref_event_url,
    create_latest_views_from_dataset,
    get_onix_records,
)
from observatory.platform.api import get_dataset_releases
from observatory.platform.observatory_config import Workflow
from observatory.platform.files import load_jsonl
from observatory.platform.bigquery import bq_find_schema, bq_run_query, bq_sharded_table_id, bq_table_id
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.config import module_file_path
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    Table,
    find_free_port,
    random_id,
    make_dummy_dag,
    bq_load_tables,
    load_and_parse_json,
)


class TestOnixWorkflow(ObservatoryTestCase):
    """Functionally test the workflow"""

    onix_data = [
        {
            "ISBN13": "111",
            "RelatedWorks": [
                {
                    "WorkRelationCode": "Manifestation of",
                    "WorkIdentifiers": [
                        {"WorkIDType": "ISBN-13", "IDValue": "112"},
                    ],
                },
                {
                    "WorkRelationCode": "Manifestation of",
                    "WorkIdentifiers": [
                        {"WorkIDType": "ISBN-13", "IDValue": "113"},
                    ],
                },
            ],
            "RelatedProducts": [
                {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "211"},
            ],
        },
        {
            "ISBN13": "112",
            "RelatedWorks": [
                {
                    "WorkRelationCode": "Manifestation of",
                    "WorkIdentifiers": [
                        {"WorkIDType": "ISBN-13", "IDValue": "112"},
                    ],
                },
            ],
            "RelatedProducts": [],
        },
        {
            "ISBN13": "211",
            "RelatedWorks": [
                {
                    "WorkRelationCode": "Manifestation of",
                    "WorkIdentifiers": [
                        {"WorkIDType": "ISBN-13", "IDValue": "211"},
                    ],
                },
            ],
            "RelatedProducts": [],
        },
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.timestamp = pendulum.now()
        self.onix_table_id = "onix"
        self.test_onix_folder = random_id()  # "onix_workflow_test_onix_table"

        # For telescope Initialisation
        self.snapshot_date = pendulum.datetime(2021, 1, 1)
        self.dag_id = "onix_workflow_test"
        self.gcp_project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.fake_onix_data_partner = OaebuPartner(  # Onix data partner to pass to the telescope for initialisation
            type_id="irus_oapen",
            bq_dataset_id="test_dataset",
            bq_table_name="test_table",
            isbn_field_name="isbn",
            title_field_name="title",
            sharded=True,
            schema_path=os.path.join(default_schema_folder(workflow_module="onix_telescope"), "onix.json"),
        )

        # fixtures folder location
        self.fixtures_folder = test_fixtures_folder(workflow_module="onix_workflow")
        # vcrpy cassettes for http request mocking
        self.events_cassette = os.path.join(self.fixtures_folder, "crossref_events_request.yaml")

    @patch("oaebu_workflows.onix_workflow.onix_workflow.bq_select_table_shard_dates")
    def test_make_release(self, mock_sel_table_suffixes):
        """Tests that the make_release function works as intended"""
        # Use a different onix snapshot date for testing purposes
        onix_snapshot_date = self.snapshot_date.add(days=1)
        crossref_snapshot_date = self.snapshot_date
        mock_sel_table_suffixes.side_effect = [[onix_snapshot_date], [crossref_snapshot_date]]
        env = ObservatoryEnvironment(
            self.gcp_project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        with env.create():
            wf = OnixWorkflow(
                dag_id="test_make_release",
                cloud_workspace=self.fake_cloud_workspace,
                data_partners=[self.fake_onix_data_partner],
                metadata_partner="onix",
            )
            dag = wf.make_dag()
            with env.create_dag_run(dag, self.snapshot_date.add(days=1)):
                release = wf.make_release(
                    data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                    run_id=env.dag_run.run_id,
                )
                self.assertEqual(release.dag_id, wf.dag_id)

                # Test release file names are as expected
                self.assertEqual(release.workslookup_path, os.path.join(release.transform_folder, "worksid.jsonl.gz"))
                self.assertEqual(
                    release.workslookup_errors_path,
                    os.path.join(release.transform_folder, "worksid_errors.jsonl.gz"),
                )
                self.assertEqual(
                    release.worksfamilylookup_path, os.path.join(release.transform_folder, "workfamilyid.jsonl.gz")
                )
                self.assertEqual(
                    release.crossref_metadata_path,
                    os.path.join(release.transform_folder, "crossref_metadata.jsonl.gz"),
                )
                self.assertEqual(
                    release.crossref_events_path, os.path.join(release.transform_folder, "crossref_events.jsonl.gz")
                )

                # Test that the onix and crossref snapshots are as expected
                self.assertEqual(onix_snapshot_date, release.onix_snapshot_date)
                self.assertEqual(crossref_snapshot_date, release.crossref_master_snapshot_date)

                # Test for case - no ONIX releases found
                mock_sel_table_suffixes.side_effect = [[]]
                with self.assertRaisesRegex(AssertionError, "ONIX"):
                    release = wf.make_release(
                        data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                        run_id=env.dag_run.run_id,
                    )

                # Test for case - no Crossref releases found
                mock_sel_table_suffixes.side_effect = [[onix_snapshot_date], []]  # No crossref releases
                with self.assertRaisesRegex(AssertionError, "Crossref"):
                    release = wf.make_release(
                        data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                        run_id=env.dag_run.run_id,
                    )

    def test_cleanup(self):
        """Tests the cleanup function of the workflow"""
        env = ObservatoryEnvironment(
            self.gcp_project_id,
            self.data_location,
            api_host="localhost",
            api_port=find_free_port(),
        )
        with env.create():
            wf = OnixWorkflow(
                dag_id="test_cleanup",
                cloud_workspace=self.fake_cloud_workspace,
                data_partners=[self.fake_onix_data_partner],
                metadata_partner="onix",
            )
            release = OnixWorkflowRelease(
                dag_id=wf.dag_id,
                run_id="test_run_id",
                snapshot_date=self.snapshot_date,
                onix_snapshot_date=self.snapshot_date,
                crossref_master_snapshot_date=self.snapshot_date,
            )
            self.assertTrue(os.path.exists(release.download_folder))
            self.assertTrue(os.path.exists(release.extract_folder))
            self.assertTrue(os.path.exists(release.transform_folder))
            wf.cleanup(release, execution_date=self.snapshot_date)
            self.assert_cleanup(release.workflow_folder)

    def test_dag_load(self):
        """Test that the DAG loads"""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="onix_workflow_test_dag_load",
                    name="Onix Workflow Test Dag Load",
                    class_name="oaebu_workflows.onix_workflow.onix_workflow.OnixWorkflow",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(
                        sensor_dag_ids=[
                            "google_analytics3_test",
                            "google_books_sales_test",
                            "google_books_traffic_test",
                            "jstor_country_test",
                            "jstor_institution_test",
                            "irus_oapen_test",
                        ],
                        data_partners=[
                            "google_analytics3",
                            "google_books_sales",
                            "google_books_traffic",
                            "jstor_country",
                            "jstor_institution",
                            "irus_oapen",
                            "nonexistent_partner",
                        ],
                        metadata_partner="onix",
                    ),
                )
            ]
        )

        with env.create() as dag_folder:
            # This should raise one error for nonexistent partner
            shutil.copy(
                os.path.join(module_file_path("observatory.platform.dags"), "load_workflows.py"), "load_workflows.py"
            )
            dag_bag = DagBag(dag_folder=dag_folder)
            self.assertNotEqual({}, dag_bag.import_errors)
            self.assertEqual(len(dag_bag.import_errors), 1)

        # Remove the nonexistent partner and onix partner
        env.workflows[0].kwargs["data_partners"] = env.workflows[0].kwargs["data_partners"][:-1]
        with env.create():
            # Should not raise any errors
            self.assert_dag_load_from_config("onix_workflow_test_dag_load")

    def test_dag_structure(self):
        """Tests that the dag structure is created as expected on dag load"""
        env = ObservatoryEnvironment(
            self.gcp_project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        with env.create():
            data_partners = [
                "jstor_country",
                "irus_oapen",
                "google_books_sales",
                "google_books_traffic",
                "google_analytics3",
            ]
            sensor_dag_ids = ["jstor", "irus_oapen", "google_books", "onix", "google_analytics3"]
            dag = OnixWorkflow(
                dag_id=self.dag_id,
                cloud_workspace=self.fake_cloud_workspace,
                data_partners=data_partners,
                metadata_partner="onix",
                sensor_dag_ids=sensor_dag_ids,
            ).make_dag()
            expected_dag_structure = {
                "google_analytics3_sensor": ["aggregate_works"],
                "google_books_sensor": ["aggregate_works"],
                "jstor_sensor": ["aggregate_works"],
                "irus_oapen_sensor": ["aggregate_works"],
                "onix_sensor": ["aggregate_works"],
                "aggregate_works": ["upload_aggregation_tables"],
                "upload_aggregation_tables": ["bq_load_aggregations"],
                "bq_load_aggregations": ["create_oaebu_crossref_metadata_table"],
                "create_oaebu_crossref_metadata_table": ["create_oaebu_crossref_events_table"],
                "create_oaebu_crossref_events_table": ["create_oaebu_book_table"],
                "create_oaebu_book_table": [
                    "create_oaebu_intermediate_table_google_books_sales",
                    "create_oaebu_intermediate_table_irus_oapen",
                    "create_oaebu_intermediate_table_google_analytics3",
                    "create_oaebu_intermediate_table_google_books_traffic",
                    "create_oaebu_intermediate_table_jstor_country",
                ],
                "create_oaebu_intermediate_table_jstor_country": ["create_oaebu_book_product_table"],
                "create_oaebu_intermediate_table_irus_oapen": ["create_oaebu_book_product_table"],
                "create_oaebu_intermediate_table_google_books_sales": ["create_oaebu_book_product_table"],
                "create_oaebu_intermediate_table_google_books_traffic": ["create_oaebu_book_product_table"],
                "create_oaebu_intermediate_table_google_analytics3": ["create_oaebu_book_product_table"],
                "create_oaebu_book_product_table": [
                    "export_oaebu_table_book_product_publisher_metrics",
                    "create_oaebu_data_qa_isbn_onix",
                    "create_oaebu_data_qa_isbn_jstor_country",
                    "create_oaebu_data_qa_eisbn_jstor_country",
                    "create_oaebu_data_qa_intermediate_unmatched_jstor_country",
                    "export_oaebu_table_book_product_metrics_country",
                    "create_oaebu_data_qa_isbn_google_books_sales",
                    "create_oaebu_data_qa_intermediate_unmatched_google_analytics3",
                    "export_oaebu_table_book_product_subject_year_metrics",
                    "export_oaebu_table_book_product_subject_thema_metrics",
                    "create_oaebu_data_qa_isbn_google_books_traffic",
                    "export_oaebu_table_book_product_subject_bic_metrics",
                    "create_oaebu_data_qa_intermediate_unmatched_google_books_sales",
                    "export_oaebu_table_book_product_list",
                    "export_oaebu_table_book_product_metrics_city",
                    "export_oaebu_table_book_product_author_metrics",
                    "export_oaebu_table_book_product_metrics_institution",
                    "create_oaebu_data_qa_intermediate_unmatched_irus_oapen",
                    "create_oaebu_data_qa_isbn_google_analytics3",
                    "export_oaebu_table_book_product_year_metrics",
                    "export_oaebu_table_book_product_subject_bisac_metrics",
                    "create_oaebu_data_qa_onix_aggregate",
                    "export_oaebu_table_book_product_metrics_events",
                    "export_oaebu_table_institution_list",
                    "create_oaebu_data_qa_intermediate_unmatched_google_books_traffic",
                    "create_oaebu_data_qa_isbn_irus_oapen",
                    "export_oaebu_table_book_product_metrics",
                ],
                "create_oaebu_data_qa_isbn_onix": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_onix_aggregate": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_isbn_jstor_country": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_eisbn_jstor_country": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_intermediate_unmatched_jstor_country": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_isbn_irus_oapen": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_intermediate_unmatched_irus_oapen": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_isbn_google_books_sales": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_intermediate_unmatched_google_books_sales": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_isbn_google_books_traffic": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_intermediate_unmatched_google_books_traffic": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_isbn_google_analytics3": ["export_oaebu_qa_metrics"],
                "create_oaebu_data_qa_intermediate_unmatched_google_analytics3": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_list": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_metrics_country": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_metrics_institution": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_institution_list": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_metrics_city": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_metrics_events": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_publisher_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_subject_bic_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_subject_bisac_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_subject_thema_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_year_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_subject_year_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_table_book_product_author_metrics": ["export_oaebu_qa_metrics"],
                "export_oaebu_qa_metrics": ["create_oaebu_latest_views"],
                "create_oaebu_latest_views": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            }
            self.assert_dag_structure(expected_dag_structure, dag)

    @patch("oaebu_workflows.onix_workflow.onix_workflow.bq_run_query")
    def test_create_and_load_aggregate_works_table(self, mock_bq_query):
        mock_bq_query.return_value = TestOnixWorkflow.onix_data
        workslookup_expected = [
            {"isbn13": "112", "work_id": "111"},
            {"isbn13": "111", "work_id": "111"},
            {"isbn13": "211", "work_id": "211"},
        ]
        workslookup_errors_expected = [
            {
                "Error": "Product ISBN13:111 is a manifestation of ISBN13:113, which is not given as a product identifier in any ONIX product record."
            }
        ]
        worksfamilylookup_expected = [
            {"isbn13": "112", "work_family_id": "111"},
            {"isbn13": "111", "work_family_id": "111"},
            {"isbn13": "211", "work_family_id": "111"},
        ]
        env = ObservatoryEnvironment(
            self.gcp_project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        with env.create():
            wf = OnixWorkflow(
                dag_id=self.dag_id,
                cloud_workspace=env.cloud_workspace,
                data_partners=[self.fake_onix_data_partner],
                bq_onix_workflow_dataset=env.add_dataset(),
                metadata_partner="onix",
            )
            dag = wf.make_dag()
            with env.create_dag_run(dag, self.snapshot_date.add(days=1)):
                release = OnixWorkflowRelease(
                    dag_id="aggregation_test",
                    run_id=env.dag_run.run_id,
                    snapshot_date=self.snapshot_date,
                    onix_snapshot_date=self.snapshot_date,
                    crossref_master_snapshot_date=self.snapshot_date,
                )
                wf.aggregate_works(release)  # Test works aggregation
                wf.upload_aggregation_tables(release)  # Test upload_aggregation_tables
                wf.bq_load_aggregations(release, ti=MagicMock(task_id=""))  # Test bq_load_workid_lookup

                ### Make Assertions ###
                self.assertTrue(os.path.exists(release.workslookup_path))
                self.assertTrue(os.path.exists(release.workslookup_errors_path))
                self.assertTrue(os.path.exists(release.worksfamilylookup_path))

                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.workslookup_path),
                    release.workslookup_path,
                )
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.workslookup_errors_path),
                    release.workslookup_errors_path,
                )
                self.assert_blob_integrity(
                    env.transform_bucket,
                    gcs_blob_name_from_path(release.worksfamilylookup_path),
                    release.worksfamilylookup_path,
                )

                table_id = bq_sharded_table_id(
                    wf.cloud_workspace.project_id,
                    wf.bq_onix_workflow_dataset,
                    wf.bq_worksid_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, len(workslookup_expected))
                self.assert_table_content(table_id, load_jsonl(release.workslookup_path), primary_key="isbn13")
                self.assert_table_content(table_id, workslookup_expected, primary_key="isbn13")

                table_id = bq_sharded_table_id(
                    wf.cloud_workspace.project_id,
                    wf.bq_onix_workflow_dataset,
                    wf.bq_worksid_error_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, len(workslookup_errors_expected))
                self.assert_table_content(table_id, load_jsonl(release.workslookup_errors_path), primary_key="Error")
                self.assert_table_content(table_id, workslookup_errors_expected, primary_key="Error")

                table_id = bq_sharded_table_id(
                    wf.cloud_workspace.project_id,
                    wf.bq_onix_workflow_dataset,
                    wf.bq_workfamilyid_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, len(worksfamilylookup_expected))
                self.assert_table_content(table_id, load_jsonl(release.worksfamilylookup_path), primary_key="isbn13")
                self.assert_table_content(table_id, worksfamilylookup_expected, primary_key="isbn13")

    def test_crossref_API_calls(self):
        """
        Test the functions that query the crossref event and metadata APIs
        """
        # Test function calls with mocked HTTP responses
        good_test_doi = "10.5555/12345678"
        bad_test_doi = "10.1111111111111"
        mailto = "agent@observatory.academy"

        with vcr.use_cassette(
            os.path.join(self.fixtures_folder, "crossref_download_function_test.yaml"), record_mode="none"
        ):
            events_start = pendulum.date(2020, 1, 1)
            events_end = pendulum.date(2022, 1, 1)
            event_url = CROSSREF_EVENT_URL_TEMPLATE.format(
                doi=good_test_doi,
                mailto=mailto,
                start_date=events_start.strftime("%Y-%m-%d"),
                end_date=events_end.strftime("%Y-%m-%d"),
            )
            events = download_crossref_event_url(event_url)
            assert events == [{"passed": True}], f"Event return incorrect. Got {events}"

        good_events = download_crossref_events([good_test_doi], events_start, events_end, mailto, max_threads=1)
        bad_events = download_crossref_events([bad_test_doi], events_start, events_end, mailto, max_threads=1)
        assert good_events, "Events should have returned something"
        assert len(good_events) == 4
        assert not bad_events, f"Events should have returned nothing, instead returned {bad_events}"

    def test_crossref_transform(self):
        """Test the function that transforms the crossref events data"""

        input_events = [
            {
                "license": "https://creativecommons.org/publicdomain/zero/1.0/",
                "obj_id": "https://doi.org/10.5555/12345678",
                "source_token": "36c35e23-8757-4a9d-aacf-345e9b7eb50d",
                "occurred_at": "2021-07-21T13:27:50Z",
                "subj_id": "https://en.wikipedia.org/api/rest_v1/page/html/Josiah_S._Carberry/1034726541",
                "id": "c76bbfd9-5122-4859-82c8-505b1eb845fd",
                "evidence_record": "https://evidence.eventdata.crossref.org/evidence/20210721-wikipedia-83789066-1794-4c4f-bb2c-ed47ccd82264",
                "terms": "https://doi.org/10.13003/CED-terms-of-use",
                "action": "add",
                "subj": {
                    "pid": "https://en.wikipedia.org/wiki/Josiah_S._Carberry",
                    "url": "https://en.wikipedia.org/w/index.php?title=Josiah_S._Carberry&oldid=1034726541",
                    "title": "Josiah S. Carberry",
                    "work_type_id": "entry-encyclopedia",
                    "api-url": "https://en.wikipedia.org/api/rest_v1/page/html/Josiah_S._Carberry/1034726541",
                },
                "source_id": "wikipedia",
                "obj": {
                    "pid": "https://doi.org/10.5555/12345678",
                    "url": "http://psychoceramics.labs.crossref.org/10.5555-12345678.html",
                    "method": "landing-page-meta-tag",
                    "verification": "checked-url-exact",
                },
                "timestamp": "2021-07-21T14:00:58Z",
                "relation_type_id": "references",
            }
        ]
        expected_transformed_events = [
            {
                "license": "https://creativecommons.org/publicdomain/zero/1.0/",
                "obj_id": "https://doi.org/10.5555/12345678",
                "source_token": "36c35e23-8757-4a9d-aacf-345e9b7eb50d",
                "occurred_at": "2021-07-21T13:27:50+00:00",
                "subj_id": "https://en.wikipedia.org/api/rest_v1/page/html/Josiah_S._Carberry/1034726541",
                "id": "c76bbfd9-5122-4859-82c8-505b1eb845fd",
                "evidence_record": "https://evidence.eventdata.crossref.org/evidence/20210721-wikipedia-83789066-1794-4c4f-bb2c-ed47ccd82264",
                "terms": "https://doi.org/10.13003/CED-terms-of-use",
                "action": "add",
                "subj": {
                    "pid": "https://en.wikipedia.org/wiki/Josiah_S._Carberry",
                    "url": "https://en.wikipedia.org/w/index.php?title=Josiah_S._Carberry&oldid=1034726541",
                    "title": "Josiah S. Carberry",
                    "work_type_id": "entry-encyclopedia",
                    "api_url": "https://en.wikipedia.org/api/rest_v1/page/html/Josiah_S._Carberry/1034726541",
                },
                "source_id": "wikipedia",
                "obj": {
                    "pid": "https://doi.org/10.5555/12345678",
                    "url": "http://psychoceramics.labs.crossref.org/10.5555-12345678.html",
                    "method": "landing-page-meta-tag",
                    "verification": "checked-url-exact",
                },
                "timestamp": "2021-07-21T14:00:58+00:00",
                "relation_type_id": "references",
            }
        ]
        # Standalone transform
        actual_transformed_events = transform_event(input_events[0])
        assert expected_transformed_events[0] == actual_transformed_events
        # List transform
        actual_transformed_events = transform_crossref_events(input_events)
        assert len(actual_transformed_events) == 1
        assert expected_transformed_events == actual_transformed_events

    def test_utility_functions(self):
        """
        Test the standalone functions in the Onix workflow that aren't specifially tested in other classes
        """

        ############################
        ### Test dois_from_table ###
        ############################

        env = ObservatoryEnvironment(
            self.gcp_project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        fake_doi_isbn_dataset_id = env.add_dataset(prefix="doi_isbn_test")
        fake_sharded_dataset = env.add_dataset(prefix="sharded_data")
        fake_view_dataset = env.add_dataset(prefix="views")
        fake_table_schema = os.path.join(default_schema_folder(workflow_module="onix_telescope"), "onix.json")
        with env.create():
            doi_table_file = os.path.join(self.fixtures_folder, "doi_isbn_query_test.jsonl")

            fake_doi_isbn_table = load_jsonl(doi_table_file)
            test_table = Table(
                "doi_isbn_test",
                False,
                fake_doi_isbn_dataset_id,
                fake_doi_isbn_table,
                fake_table_schema,  # use onix schema just for ease uploading
            )

            release_date = pendulum.datetime(2022, 6, 13)
            bq_load_tables(
                tables=[test_table],
                bucket_name=env.transform_bucket,
                snapshot_date=release_date,
                project_id=self.gcp_project_id,
            )
            table_id = bq_table_id(self.gcp_project_id, fake_doi_isbn_dataset_id, "doi_isbn_test")
            actual_dois = dois_from_table(table_id, doi_column_name="DOI", distinct=True)
            fake_doi_isbns = [entry["DOI"] for entry in fake_doi_isbn_table]

            # Check there are no duplicates and the contents are the same
            assert len(actual_dois) == len(set(fake_doi_isbns))
            assert set(actual_dois) == set(fake_doi_isbns)

            # Do the same but allow duplicates
            actual_dois = dois_from_table(table_id, doi_column_name="DOI", distinct=False)
            fake_doi_isbns = [entry["DOI"] for entry in fake_doi_isbn_table]
            assert len(actual_dois) == len(fake_doi_isbns)
            assert sorted(actual_dois) == sorted(fake_doi_isbns)

            #############################################
            ### Test create_latest_views_from_dataset ###
            #############################################

            # We will use the onix.jsonl fixture for testing.
            previous_release_date = pendulum.datetime(2022, 6, 6)  # Add another release date
            data = [fake_doi_isbn_table, fake_doi_isbn_table[:2]]
            for release, table in zip([release_date, previous_release_date], data):
                test_table = Table("data_export", True, fake_sharded_dataset, table, fake_table_schema)
                bq_load_tables(
                    tables=[test_table],
                    bucket_name=env.transform_bucket,
                    snapshot_date=release,
                    project_id=self.gcp_project_id,
                )

            # Now make the views from this dataset
            create_latest_views_from_dataset(
                project_id=self.gcp_project_id,
                from_dataset=fake_sharded_dataset,
                to_dataset=fake_view_dataset,
                date_match=release_date.strftime("%Y%m%d"),
                data_location=self.data_location,
            )

            # Grab the data from the view and make assertions
            view_data = bq_run_query(f"SELECT * FROM {self.gcp_project_id}.{fake_view_dataset}.data_export")
            view_isbns = [entry["ISBN13"] for entry in view_data]
            view_dois = [entry["DOI"] for entry in view_data]
            actual_isbns = [entry["ISBN13"] for entry in fake_doi_isbn_table]
            actual_dois = [entry["DOI"] for entry in fake_doi_isbn_table]

            assert len(view_data) == len(fake_doi_isbn_table)
            assert len(actual_isbns) == len(view_isbns)
            assert sorted(actual_isbns) == sorted(view_isbns)
            assert len(actual_dois) == len(view_dois)
            assert sorted(actual_dois) == sorted(view_dois)

    def setup_fake_lookup_tables(
        self, settings_dataset_id: str, fixtures_dataset_id: str, release_date: pendulum.DateTime, bucket_name: str
    ):
        """Create a new onix and subject lookup and country tables with their own dataset and table ids.
        Populate them with some fake data.

        :param settings_dataset_id: The dataset to store the country table
        :param fixtures_dataset_id: The dataset to store the lookup tables
        :param release_date: The release/snapshot date
        :param bucket_name: The GCP bucket name to load the data to
        """

        table_and_dataset = [
            ("bic_lookup", fixtures_dataset_id),
            ("bisac_lookup", fixtures_dataset_id),
            ("thema_lookup", fixtures_dataset_id),
            ("country", settings_dataset_id),
        ]

        # Create the table objects
        onix_test_schema_folder = os.path.join(self.fixtures_folder, "schema")
        tables = []
        for table_name, dataset in table_and_dataset:
            tables.append(
                Table(
                    table_name,
                    False,
                    dataset,
                    load_jsonl(os.path.join(self.fixtures_folder, "e2e_inputs", f"{table_name}.jsonl")),
                    bq_find_schema(path=onix_test_schema_folder, table_name=table_name),
                )
            )

        # Load tables into bigquery
        bq_load_tables(
            tables=tables, bucket_name=bucket_name, snapshot_date=release_date, project_id=self.gcp_project_id
        )

    def setup_input_data(
        self,
        settings_dataset_id: str,
        fixtures_dataset_id: str,
        crossref_master_dataset_id: str,
        partner_dataset: str,
        onix_dataset_id: str,
        release_date: pendulum.DateTime,
        bucket_name: str,
        include_google_analytics3: bool,
    ) -> List[OaebuPartner]:
        """Uploads the data partner fixtures to their respective GCP bucket and bigquery tables.
        Creates the partners based on the originals - but changes the dataset ids for tests.
        Create a new onix and subject lookup and country tables with their own dataset and table ids.

        :param settings_dataset_id: The dataset to store the country table
        :param fixtures_dataset_id: The dataset to store the lookup tables
        :param partner_dataset: The bigquery dataset ID to load the data partner tables to
        :param crossref_master_dataset_id: The bigquery dataset ID of the master crossref table
        :param onix_dataaset: The Bigquery dataset ID to load the onix partner table to
        :param release_date: The release/snapshot date of sharded tables
        :param bucket_name: The name of the bucket to upload the jsonl files to
        :param include_google_analytics3: Whether to include google analytics 3 as a partner
        :return: The resulting OaebuPartners
        """

        ############################################
        ### Upload lookups, country and crossref ###
        ############################################

        table_and_dataset = [
            ("bic_lookup", fixtures_dataset_id, False),
            ("bisac_lookup", fixtures_dataset_id, False),
            ("thema_lookup", fixtures_dataset_id, False),
            ("country", settings_dataset_id, False),
            ("crossref_metadata_master", crossref_master_dataset_id, True),
        ]

        # Create the table objects
        onix_test_schema_folder = os.path.join(self.fixtures_folder, "schema")
        tables = []
        for table_name, dataset, sharded in table_and_dataset:
            tables.append(
                Table(
                    table_name,
                    sharded,
                    dataset,
                    load_jsonl(os.path.join(self.fixtures_folder, "e2e_inputs", f"{table_name}.jsonl")),
                    bq_find_schema(path=onix_test_schema_folder, table_name=table_name),
                )
            )

        ###########################
        ### Upload partner data ###
        ###########################

        table_dataset_sharded_schema = [
            ("jstor_country", partner_dataset),
            ("jstor_institution", partner_dataset),
            ("google_books_sales", partner_dataset),
            ("google_books_traffic", partner_dataset),
            ("irus_oapen", partner_dataset),
            ("irus_fulcrum", partner_dataset),
            ("ucl_discovery", partner_dataset),
        ]
        if include_google_analytics3:
            table_dataset_sharded_schema.append(("google_analytics3", partner_dataset))

        # Create partner objects define tables
        partners = []
        for partner_name, dataset in table_dataset_sharded_schema:
            partner = partner_from_str(partner_name)
            partner.bq_dataset_id = dataset
            partners.append(partner)
            tables.append(
                Table(
                    partner_name,
                    partner.sharded,
                    partner.bq_dataset_id,
                    load_jsonl(os.path.join(self.fixtures_folder, "e2e_inputs", f"{partner_name}.jsonl")),
                    partner.schema_path,
                )
            )

        # Create metadata table
        metadata_partner = partner_from_str("onix", metadata_partner=True)
        metadata_partner.bq_dataset_id = onix_dataset_id
        tables.append(
            Table(
                "onix",
                metadata_partner.sharded,
                metadata_partner.bq_dataset_id,
                load_jsonl(os.path.join(self.fixtures_folder, "e2e_inputs", f"onix.jsonl")),
                metadata_partner.schema_path,
            )
        )

        # Load tables into bigquery
        bq_load_tables(
            tables=tables, bucket_name=bucket_name, snapshot_date=release_date, project_id=self.gcp_project_id
        )

        return partners, metadata_partner

    def run_telescope_tests(self, *, include_google_analytics3: bool = False):
        """Functional test of the ONIX workflow"""

        def vcr_ignore_condition(request):
            """This function is used by vcrpy to allow requests to sources not in the cassette file.
            At time of writing, the only mocked requests are the ones to crossref events."""
            allowed_domains = ["https://api.eventdata.crossref.org"]
            allow_request = any([request.url.startswith(i) for i in allowed_domains])
            if not allow_request:
                return None
            return request

        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.gcp_project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )

        # Create workflow datasets
        onix_workflow_dataset_id = env.add_dataset(prefix="onix_workflow")
        master_crossref_dataset_id = env.add_dataset(prefix="crossref_master")
        oaebu_data_qa_dataset_id = env.add_dataset(prefix="oaebu_data_qa")
        oaebu_latest_data_qa_dataset_id = env.add_dataset(prefix="oaebu_data_qa_latest")
        oaebu_intermediate_dataset_id = env.add_dataset(prefix="oaebu_intermediate")
        oaebu_output_dataset_id = env.add_dataset(prefix="oaebu_output")
        oaebu_export_dataset_id = env.add_dataset(prefix="oaebu_export")
        oaebu_latest_export_dataset_id = env.add_dataset(prefix="oaebu_export_latest")
        oaebu_settings_dataset_id = env.add_dataset(prefix="settings")
        oaebu_fixtures_dataset_id = env.add_dataset(prefix="fixtures")
        oaebu_crossref_dataset_id = env.add_dataset(prefix="crossref")

        # Fake partner datasets
        onix_dataset_id = env.add_dataset(prefix="onix")
        fake_partner_dataset_id = env.add_dataset(prefix="partner")

        # Create the Observatory environment and run tests
        with env.create(task_logging=True):
            # Setup data partners, remove Google Analytics (the last partner) from these tests
            partner_release_date = pendulum.datetime(2021, 5, 15)
            data_partners, metadata_partner = self.setup_input_data(
                settings_dataset_id=oaebu_settings_dataset_id,
                fixtures_dataset_id=oaebu_fixtures_dataset_id,
                crossref_master_dataset_id=master_crossref_dataset_id,
                partner_dataset=fake_partner_dataset_id,
                onix_dataset_id=onix_dataset_id,
                release_date=partner_release_date,
                bucket_name=env.transform_bucket,
                include_google_analytics3=include_google_analytics3,
            )

            # Expected sensor dag_ids
            sensor_dag_ids = [
                "jstor",
                "google_books",
                "google_analytics3",
                "irus_oapen",
                "irus_fulcrum",
                "ucl_discovery",
                "onix",
            ]

            # Setup telescope
            telescope = OnixWorkflow(
                dag_id=f"onix_workflow_test",
                cloud_workspace=env.cloud_workspace,
                metadata_partner=metadata_partner,
                bq_master_crossref_project_id=env.cloud_workspace.project_id,
                bq_master_crossref_dataset_id=master_crossref_dataset_id,
                bq_oaebu_crossref_dataset_id=oaebu_crossref_dataset_id,
                bq_master_crossref_metadata_table_name="crossref_metadata_master",  # Set in setup_input_data()
                bq_country_project_id=env.cloud_workspace.project_id,
                bq_country_dataset_id=oaebu_settings_dataset_id,
                bq_subject_project_id=env.cloud_workspace.project_id,
                bq_subject_dataset_id=oaebu_fixtures_dataset_id,
                bq_oaebu_data_qa_dataset=oaebu_data_qa_dataset_id,
                bq_oaebu_latest_data_qa_dataset=oaebu_latest_data_qa_dataset_id,
                bq_onix_workflow_dataset=onix_workflow_dataset_id,
                bq_oaebu_intermediate_dataset=oaebu_intermediate_dataset_id,
                bq_oaebu_dataset=oaebu_output_dataset_id,
                bq_oaebu_export_dataset=oaebu_export_dataset_id,
                bq_oaebu_latest_export_dataset=oaebu_latest_export_dataset_id,
                data_partners=data_partners,
                sensor_dag_ids=sensor_dag_ids,
                start_date=pendulum.datetime(year=2021, month=5, day=9),
                crossref_start_date=pendulum.datetime(year=2018, month=5, day=14),
                max_threads=1,  # Use 1 thread for tests
            )

            # Skip dag existence check in sensor.
            for sensor in telescope.operators[0]:
                sensor.check_exists = False
                sensor.grace_period = timedelta(seconds=1)
            workflow_dag = telescope.make_dag()

            # If there is no dag run in the search interval, sensor will return success.
            expected_state = "success"
            with env.create_dag_run(workflow_dag, telescope.start_date):
                for sensor_id in sensor_dag_ids:
                    ti = env.run_task(f"{sensor_id}_sensor")
                    self.assertEqual(ti.state, State.SUCCESS)

            # Run Dummy Dags
            execution_date = pendulum.datetime(year=2021, month=5, day=16)
            for dag_id in sensor_dag_ids:
                dag = make_dummy_dag(dag_id, execution_date)
                with env.create_dag_run(dag, execution_date):
                    # Running all of a DAGs tasks sets the DAG to finished
                    ti = env.run_task("dummy_task")
                    self.assertEqual(ti.state, State.SUCCESS)

            # Run end to end tests for DOI DAG
            with env.create_dag_run(workflow_dag, execution_date):
                # Test that sensors go into 'success' state as the DAGs that they are waiting for have finished
                for dag_id in sensor_dag_ids:
                    ti = env.run_task(f"{dag_id}_sensor")
                    self.assertEqual(ti.state, State.SUCCESS)

                # Mock make_release
                release_date = pendulum.datetime(year=2021, month=5, day=22)
                telescope.make_release = MagicMock(
                    return_value=OnixWorkflowRelease(
                        dag_id=telescope.dag_id,
                        run_id=env.dag_run.run_id,
                        snapshot_date=release_date,
                        onix_snapshot_date=partner_release_date,
                        crossref_master_snapshot_date=partner_release_date,
                    )
                )
                release = telescope.make_release()

                # Make the SQL queries for data QA
                release_suffix = release.snapshot_date.strftime("%Y%m%d")
                data_qa_sql = {
                    "onix_invalid_isbn": f"SELECT ISBN13 from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.onix_invalid_isbn{release_suffix}",
                    "onix_aggregate_metrics": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.onix_aggregate_metrics{release_suffix}",
                    "jstor_country_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_invalid_isbn{release_suffix}",
                    "jstor_country_invalid_eisbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_invalid_eisbn{release_suffix}",
                    "jstor_country_unmatched_ISBN": f"SELECT ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_country_unmatched_ISBN{release_suffix}",
                    "jstor_institution_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_invalid_isbn{release_suffix}",
                    "jstor_institution_invalid_eisbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_invalid_eisbn{release_suffix}",
                    "jstor_institution_unmatched_ISBN": f"SELECT ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.jstor_institution_unmatched_ISBN{release_suffix}",
                    "google_books_sales_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_books_sales_invalid_isbn{release_suffix}",
                    "google_books_sales_unmatched_Primary_ISBN": f"SELECT Primary_ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_books_sales_unmatched_Primary_ISBN{release_suffix}",
                    "google_books_traffic_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_books_traffic_invalid_isbn{release_suffix}",
                    "google_books_traffic_unmatched_Primary_ISBN": f"SELECT Primary_ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_books_traffic_unmatched_Primary_ISBN{release_suffix}",
                    "irus_oapen_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.irus_oapen_invalid_isbn{release_suffix}",
                    "irus_oapen_unmatched_ISBN": f"SELECT ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.irus_oapen_unmatched_ISBN{release_suffix}",
                    "irus_fulcrum_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.irus_fulcrum_invalid_isbn{release_suffix}",
                    "irus_fulcrum_unmatched_ISBN": f"SELECT ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.irus_fulcrum_unmatched_ISBN{release_suffix}",
                    "ucl_discovery_invalid_isbn": f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.ucl_discovery_invalid_isbn{release_suffix}",
                    "ucl_discovery_unmatched_ISBN": f"SELECT ISBN from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.ucl_discovery_unmatched_ISBN{release_suffix}",
                }
                if include_google_analytics3:
                    data_qa_sql[
                        "google_analytics3_invalid_isbn"
                    ] = f"SELECT * from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_analytics3_invalid_isbn{release_suffix}"

                    data_qa_sql[
                        "google_analytics3_unmatched_publication_id"
                    ] = f"SELECT publication_id from {self.gcp_project_id}.{oaebu_data_qa_dataset_id}.google_analytics3_unmatched_publication_id{release_suffix}"

                # Create the latest data QA SQL as well
                latest_data_qa_sql = {}
                for k, sql in data_qa_sql.items():
                    latest_data_qa_sql[k] = sql.replace(
                        oaebu_data_qa_dataset_id, oaebu_latest_data_qa_dataset_id
                    ).replace(release_suffix, "")

                # Aggregate works
                ti = env.run_task(telescope.aggregate_works.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Upload aggregation tables
                ti = env.run_task(telescope.upload_aggregation_tables.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Load work id, work id errors and work family id tables into bigquery
                ti = env.run_task(telescope.bq_load_aggregations.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, onix_workflow_dataset_id, telescope.bq_worksid_table_name, release_date
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(os.path.join(self.fixtures_folder, "e2e_outputs", "onix_workid_isbn.json")),
                    primary_key="isbn13",
                )
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, onix_workflow_dataset_id, telescope.bq_worksid_error_table_name, release_date
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(
                        os.path.join(self.fixtures_folder, "e2e_outputs", "onix_workid_isbn_errors.json")
                    ),
                    primary_key="Error",
                )
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, onix_workflow_dataset_id, telescope.bq_workfamilyid_table_name, release_date
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(
                        os.path.join(self.fixtures_folder, "e2e_outputs", "onix_workfamilyid_isbn.json")
                    ),
                    primary_key="isbn13",
                )

                # Load crossref metadata table into bigquery
                ti = env.run_task(telescope.create_oaebu_crossref_metadata_table.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Assertions
                table_id = bq_sharded_table_id(
                    self.gcp_project_id,
                    oaebu_crossref_dataset_id,
                    telescope.bq_oaebu_crossref_metadata_table_name,
                    release_date,
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(os.path.join(self.fixtures_folder, "e2e_outputs", "crossref_metadata.json")),
                    primary_key="DOI",
                )

                # Load crossref event table into bigquery
                with vcr.use_cassette(
                    self.events_cassette, record_mode="none", before_record_request=vcr_ignore_condition
                ):
                    ti = env.run_task(telescope.create_oaebu_crossref_events_table.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_sharded_table_id(
                    self.gcp_project_id,
                    oaebu_crossref_dataset_id,
                    telescope.bq_crossref_events_table_name,
                    release_date,
                )
                crossref_fixture_table = load_and_parse_json(
                    os.path.join(self.fixtures_folder, "e2e_outputs", "crossref_events.json"),
                    timestamp_fields=["timestamp", "occurred_at", "updated_date", "issued", "dateModified"],
                )
                self.assert_table_content(table_id, crossref_fixture_table, primary_key="id")

                # Create book table in bigquery
                ti = env.run_task(telescope.create_oaebu_book_table.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, oaebu_output_dataset_id, telescope.bq_book_table_name, release_date
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(os.path.join(self.fixtures_folder, "e2e_outputs", "book.json")),
                    primary_key="isbn",
                )

                # Create oaebu intermediate tables - onix doesn't have an intermediate table so skip it
                for data_partner in data_partners:
                    ti = env.run_task(f"create_oaebu_intermediate_table_{data_partner.bq_table_name}")
                    self.assertEqual(ti.state, State.SUCCESS)

                # Create oaebu output tables
                ti = env.run_task(telescope.create_oaebu_book_product_table.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                fixture_name = "book_product_ga.json" if include_google_analytics3 else "book_product.json"
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, oaebu_output_dataset_id, telescope.bq_book_product_table_name, release_date
                )
                self.assert_table_content(
                    table_id,
                    load_and_parse_json(
                        os.path.join(self.fixtures_folder, "e2e_outputs", fixture_name),
                        date_fields={"month", "published_date"},
                    ),
                    primary_key="ISBN13",
                )

                ############################################################
                ### Check the QA + unmatched tasks for each data partner ###
                ############################################################

                # Check ONIX invalid ISBN13s picked up in ONIX
                ti = env.run_task(telescope.create_oaebu_data_qa_isbn_onix.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                records = bq_run_query(data_qa_sql["onix_invalid_isbn"])
                isbns = set([record["ISBN13"] for record in records])
                self.assertEqual(len(isbns), 1)
                self.assertIn("1111111111111", isbns)

                # ONIX aggregate metrics and check aggregate metrics are correct
                ti = env.run_task(telescope.create_oaebu_data_qa_onix_aggregate.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                records = bq_run_query(data_qa_sql["onix_aggregate_metrics"])
                self.assertEqual(len(records), 1)
                self.assertEqual(records[0]["table_size"], 4)
                self.assertEqual(records[0]["no_isbns"], 0)
                self.assertEqual(records[0]["no_relatedworks"], 0)
                self.assertEqual(records[0]["no_relatedproducts"], 3)
                self.assertEqual(records[0]["no_doi"], 4)
                self.assertEqual(records[0]["no_productform"], 0)
                self.assertEqual(records[0]["no_contributors"], 2)
                self.assertEqual(records[0]["no_titledetails"], 4)
                self.assertEqual(records[0]["no_publisher_urls"], 4)

                task_id_qa_isbn_fmt = "create_oaebu_data_qa_isbn_{data_partner_table_name}"
                task_id_qa_eisbn_fmt = "create_oaebu_data_qa_eisbn_{data_partner_table_name}"
                task_id_qa_unmatched_fmt = "create_oaebu_data_qa_intermediate_unmatched_{data_partner_table_name}"
                invalid_isbns_expected = set(sorted(["111", "1111111111111", "113", "211"]))
                invalid_eisbns_expected = set(
                    sorted(["111", "1111111111111", "113", None], key=lambda x: x or "0")
                )  # lambda method allows sorting a None
                unmatched_isbns_expected = set(sorted(["111", "113", "211"]))

                for partner in data_partners:
                    # Run ISBN QA task
                    ti = env.run_task(task_id_qa_isbn_fmt.format(data_partner_table_name=partner.bq_table_name))
                    self.assertEqual(ti.state, State.SUCCESS)
                    # Run unatched QA task
                    ti = env.run_task(task_id_qa_unmatched_fmt.format(data_partner_table_name=partner.bq_table_name))
                    self.assertEqual(ti.state, State.SUCCESS)

                    if partner.type_id.startswith("google_analytics3"):  # GA is a special case
                        # Check ISBNs are valid
                        records = bq_run_query(data_qa_sql["google_analytics3_invalid_isbn"])
                        isbns = set([record["publication_id"] for record in records])
                        self.assertEqual(len(isbns), 1)
                        self.assertTrue("(none)" in isbns)
                        # Check unmatched ISBNs picked up
                        records = bq_run_query(data_qa_sql["google_analytics3_unmatched_publication_id"])
                        isbns = set([record["publication_id"] for record in records])
                        self.assertEqual(len(isbns), 1)
                        self.assertTrue("(none)" in isbns)
                    else:
                        # Check ISBNs are valid
                        records = bq_run_query(data_qa_sql[f"{partner.type_id}_invalid_isbn"])
                        isbns = set(sorted([record[partner.isbn_field_name] for record in records]))
                        self.assertEqual(isbns, invalid_isbns_expected)
                        # Check unmatched ISBNs picked up
                        records = bq_run_query(
                            data_qa_sql[f"{partner.bq_table_name}_unmatched_{partner.isbn_field_name}"]
                        )
                        isbns = set(sorted([record[partner.isbn_field_name] for record in records]))
                        self.assertEqual(isbns, unmatched_isbns_expected)

                    if partner.type_id.startswith("jstor"):
                        # Run eISBN task - JSTOR is the only partner that uses eISBN
                        ti = env.run_task(task_id_qa_eisbn_fmt.format(data_partner_table_name=partner.bq_table_name))
                        self.assertEqual(ti.state, State.SUCCESS)
                        # Check eISBNs are valid
                        records = bq_run_query(data_qa_sql[f"{partner.type_id}_invalid_eisbn"])
                        eisbns = set(sorted([record["eISBN"] for record in records], key=lambda x: x or "0"))
                        self.assertEqual(eisbns, invalid_eisbns_expected)

                ##########################
                ### Export OAEBU Tables ##
                ##########################

                export_tables = [
                    ("book_product_list", 4),
                    ("book_product_metrics", 5),
                    ("book_product_metrics_country", 22),
                    ("book_product_metrics_institution", 1),
                    ("institution_list", 1),
                    ("book_product_metrics_city", 26),
                    ("book_product_metrics_events", 3),
                    ("book_product_publisher_metrics", 4),
                    ("book_product_subject_bic_metrics", 0),
                    ("book_product_subject_bisac_metrics", 0),
                    ("book_product_subject_thema_metrics", 0),
                    ("book_product_year_metrics", 5),
                    ("book_product_subject_year_metrics", 0),
                    ("book_product_author_metrics", 3),
                ]

                export_prefix = self.gcp_project_id.replace("-", "_")
                for table, exp_rows in export_tables:
                    ti = env.run_task(f"export_oaebu_table_{table}")
                    self.assertEqual(ti.state, State.SUCCESS)
                    self.assertEqual(expected_state, ti.state, msg=f"table: {table}")
                    # Check that the data_export tables tables exist and have the correct number of rows
                    table_id = bq_sharded_table_id(
                        self.gcp_project_id, oaebu_export_dataset_id, f"{export_prefix}_{table}", release_date
                    )
                    self.assert_table_integrity(table_id, expected_rows=exp_rows)

                # Book product list content assertion
                table_id = bq_sharded_table_id(
                    self.gcp_project_id, oaebu_export_dataset_id, f"{export_prefix}_book_product_list", release_date
                )
                fixture_table = load_and_parse_json(
                    os.path.join(self.fixtures_folder, "e2e_outputs", "book_product_list.json"),
                    date_fields=["published_date"],
                )
                self.assert_table_content(table_id, fixture_table, primary_key="product_id")

                # Export oaebu qa table
                ti = env.run_task(telescope.export_oaebu_qa_metrics.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                ################################
                ### Validate the joins worked ##
                ################################

                # JSTOR
                sql = f"SELECT ISBN, work_id, work_family_id from {self.gcp_project_id}.{oaebu_intermediate_dataset_id}.jstor_country_matched{release_suffix}"
                records = bq_run_query(sql)
                oaebu_works = {record["ISBN"]: record["work_id"] for record in records}
                oaebu_wfam = {record["ISBN"]: record["work_family_id"] for record in records}

                self.assertTrue(
                    oaebu_works["1111111111111"] == oaebu_works["2222222222222"]
                    and oaebu_works["113"] is None
                    and oaebu_works["111"] is None
                    and oaebu_works["211"] is None
                )

                self.assertTrue(
                    oaebu_wfam["1111111111111"] == oaebu_wfam["2222222222222"]
                    and oaebu_wfam["113"] is None
                    and oaebu_wfam["111"] is None
                    and oaebu_wfam["211"] is None
                )

                # IRUS OAPEN
                sql = f"SELECT ISBN, work_id, work_family_id from {self.gcp_project_id}.{oaebu_intermediate_dataset_id}.irus_oapen_matched{release_suffix}"
                records = bq_run_query(sql)
                oaebu_works = {record["ISBN"]: record["work_id"] for record in records}
                oaebu_wfam = {record["ISBN"]: record["work_family_id"] for record in records}

                self.assertTrue(
                    oaebu_works["1111111111111"] == oaebu_works["2222222222222"]
                    and oaebu_works["113"] is None
                    and oaebu_works["111"] is None
                    and oaebu_works["211"] is None
                )

                self.assertTrue(
                    oaebu_wfam["1111111111111"] == oaebu_wfam["2222222222222"]
                    and oaebu_wfam["113"] is None
                    and oaebu_wfam["111"] is None
                    and oaebu_wfam["211"] is None
                )

                #################################
                ### Create and validate views ###
                #################################

                ti = env.run_task(telescope.create_oaebu_latest_views.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Check data QA views are the same as the tables
                for k in data_qa_sql.keys():
                    table_data = bq_run_query(data_qa_sql[k])
                    view_data = bq_run_query(latest_data_qa_sql[k])
                    self.assertEqual(table_data, view_data)

                for export_table, exp_rows in export_tables:
                    export_view = bq_run_query(
                        f"SELECT * FROM {self.gcp_project_id}.{oaebu_latest_export_dataset_id}.{self.gcp_project_id.replace('-', '_')}_{export_table}"
                    )
                    self.assertEqual(expected_state, ti.state, msg=f"table: {table}")
                    # Check that the data_export tables (views) has the correct number of rows
                    self.assertEqual(len(export_view), exp_rows)

                ################################
                ### Add releases and Cleanup ###
                ################################

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test cleanup
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    def test_telescope(self):
        """Test that ONIX Workflow runs when Google Analytics is not included"""

        self.run_telescope_tests(include_google_analytics3=False)

    def test_telescope_with_google_analytics(self):
        """Test that ONIX Workflow runs when Google Analytics is included"""

        self.run_telescope_tests(include_google_analytics3=True)

    @patch("oaebu_workflows.onix_workflow.onix_workflow.bq_run_query")
    def test_get_onix_records(self, mock_bq_query):
        mock_bq_query.return_value = TestOnixWorkflow.onix_data
        records = get_onix_records("test_table_id")
        self.assertEqual(len(records), 3)
        self.assertEqual(records[0]["ISBN13"], "111")



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import os
import shutil
from collections import defaultdict
from unittest.mock import patch

import pendulum
from airflow.exceptions import AirflowException
from airflow.models.connection import Connection
from airflow.utils.state import State
from click.testing import CliRunner

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.google_books_telescope.google_books_telescope import (
    GoogleBooksRelease,
    GoogleBooksTelescope,
)
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    SftpServer,
    find_free_port,
    random_id,
)
from observatory.platform.bigquery import bq_table_id
from observatory.platform.observatory_config import Workflow
from observatory.platform.sftp import SftpFolders
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.api import get_dataset_releases


class TestGoogleBooksTelescope(ObservatoryTestCase):
    """Tests for the GoogleBooks telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.
        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestGoogleBooksTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.sftp_port = find_free_port()

    def test_dag_structure(self):
        """Test that the Google Books DAG has the correct structure."""
        dag = GoogleBooksTelescope(
            dag_id="test_dag", cloud_workspace=self.fake_cloud_workspace, sftp_root="/"
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["list_release_info"],
                "list_release_info": ["move_files_to_in_progress"],
                "move_files_to_in_progress": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["move_files_to_finished"],
                "move_files_to_finished": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Google Books DAG can be loaded from a DAG bag."""
        # Run tests both for telescope with file suffixes and without
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="google_books",
                    name="My Google Books Telescope",
                    class_name="oaebu_workflows.google_books_telescope.google_books_telescope.GoogleBooksTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                )
            ]
        )
        with env.create():
            self.assert_dag_load_from_config("google_books")

    def test_telescope(self):
        """Test the Google Books telescope end to end."""
        fixtures_folder = test_fixtures_folder(workflow_module="google_books_telescope")
        params = {
            "no_download_files": 2,
            "bq_rows": 4,
            "traffic_download_hash": "db4dca44d5231e0c4e2ad95db41b79b6",
            "traffic_transform_hash": "b8073007",
            "sales_download_hash": "6496518be1ea73694d0a8f89c0b42f20",
            "sales_transform_hash": "ebe49987",
            "test_files": {
                "GoogleBooksTrafficReport_2020_02.csv": os.path.join(
                    fixtures_folder, "GoogleBooksTrafficReport_2020_02.csv"
                ),
                "GoogleSalesTransactionReport_2020_02.csv": os.path.join(
                    fixtures_folder, "GoogleSalesTransactionReport_2020_02.csv"
                ),
            },
        }

        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        sftp_server = SftpServer(host="localhost", port=self.sftp_port)
        dataset_id = env.add_dataset()

        # Create the Observatory environment and run tests
        with env.create():
            with sftp_server.create() as sftp_root:
                # Setup Telescope
                execution_date = pendulum.datetime(year=2021, month=3, day=31)
                sales_partner = partner_from_str("google_books_sales")
                sales_partner.bq_dataset_id = dataset_id
                traffic_partner = partner_from_str("google_books_traffic")
                traffic_partner.bq_dataset_id = dataset_id
                telescope = GoogleBooksTelescope(
                    dag_id="google_books_test",
                    cloud_workspace=env.cloud_workspace,
                    sftp_root="/",
                    sales_partner=sales_partner,
                    traffic_partner=traffic_partner,
                )
                dag = telescope.make_dag()

                # Add SFTP connection
                conn = Connection(
                    conn_id=telescope.sftp_service_conn_id, uri=f"ssh://:password@localhost:{self.sftp_port}"
                )
                env.add_connection(conn)
                with env.create_dag_run(dag, execution_date):
                    # Test that all dependencies are specified: no error should be thrown
                    ti = env.run_task(telescope.check_dependencies.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                    # Add file to SFTP server
                    local_sftp_folders = SftpFolders(telescope.dag_id, telescope.sftp_service_conn_id, sftp_root)
                    os.makedirs(local_sftp_folders.upload, exist_ok=True)
                    for file_name, file_path in params["test_files"].items():
                        upload_file = os.path.join(local_sftp_folders.upload, file_name)
                        shutil.copy(file_path, upload_file)

                    # Check that the correct release info is returned via Xcom
                    ti = env.run_task(telescope.list_release_info.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    release_info = ti.xcom_pull(
                        key=GoogleBooksTelescope.RELEASE_INFO,
                        task_ids=telescope.list_release_info.__name__,
                        include_prior_dates=False,
                    )

                    # Get release info from SFTP server and create expected release info
                    expected_release_info = defaultdict(list)
                    for file_name, file_path in params["test_files"].items():
                        expected_release_date = pendulum.from_format(file_name[-11:].strip(".csv"), "YYYY_MM").end_of(
                            "month"
                        )
                        release_date_str = expected_release_date.format("YYYYMMDD")
                        if release_date_str == "20200229":
                            expected_release_file = os.path.join(telescope.sftp_folders.in_progress, file_name)
                            expected_release_info[release_date_str].append(expected_release_file)
                    self.assertTrue(1, len(release_info))
                    self.assertEqual(expected_release_info["20200229"].sort(), release_info["20200229"].sort())

                    # Use release info for other tasks
                    releases = []
                    for release_date, sftp_files in release_info.items():
                        releases.append(
                            GoogleBooksRelease(
                                dag_id=telescope.dag_id,
                                run_id=env.dag_run.run_id,
                                partition_date=pendulum.parse(release_date),
                                sftp_files=sftp_files,
                            )
                        )
                    self.assertTrue(1, len(releases))
                    release = releases[0]

                    # Test move file to in progress
                    ti = env.run_task(telescope.move_files_to_in_progress.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    for file in release.sftp_files:
                        file_name = os.path.basename(file)
                        upload_file = os.path.join(local_sftp_folders.upload, file_name)
                        self.assertFalse(os.path.isfile(upload_file))
                        in_progress_file = os.path.join(local_sftp_folders.in_progress, file_name)
                        self.assertTrue(os.path.isfile(in_progress_file))

                    # Run main telescope tasks
                    ti = env.run_task(telescope.download.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    ti = env.run_task(telescope.upload_downloaded.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    ti = env.run_task(telescope.transform.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    ti = env.run_task(telescope.upload_transformed.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    ti = env.run_task(telescope.bq_load.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                    # Make assertions for the above tasks
                    # Test download
                    self.assertTrue(os.path.exists(release.download_traffic_path))
                    self.assertTrue(os.path.exists(release.download_sales_path))
                    self.assert_file_integrity(release.download_traffic_path, params["traffic_download_hash"], "md5")
                    self.assert_file_integrity(release.download_sales_path, params["sales_download_hash"], "md5")

                    # Test upload downloaded
                    self.assert_blob_integrity(
                        env.download_bucket,
                        gcs_blob_name_from_path(release.download_traffic_path),
                        release.download_traffic_path,
                    )
                    self.assert_blob_integrity(
                        env.download_bucket,
                        gcs_blob_name_from_path(release.download_sales_path),
                        release.download_sales_path,
                    )

                    # Test that file transformed
                    self.assertTrue(os.path.exists(release.transform_sales_path))
                    self.assertTrue(os.path.exists(release.transform_traffic_path))
                    self.assert_file_integrity(release.transform_sales_path, params["sales_transform_hash"], "gzip_crc")
                    self.assert_file_integrity(
                        release.transform_traffic_path, params["traffic_transform_hash"], "gzip_crc"
                    )

                    # Test that transformed file uploaded
                    self.assert_blob_integrity(
                        env.transform_bucket,
                        gcs_blob_name_from_path(release.transform_traffic_path),
                        release.transform_traffic_path,
                    )
                    self.assert_blob_integrity(
                        env.transform_bucket,
                        gcs_blob_name_from_path(release.transform_sales_path),
                        release.transform_sales_path,
                    )

                    # Test that data loaded into BigQuery
                    table_id = bq_table_id(
                        telescope.cloud_workspace.project_id,
                        telescope.sales_partner.bq_dataset_id,
                        telescope.sales_partner.bq_table_name,
                    )
                    self.assert_table_integrity(table_id, params["bq_rows"])
                    table_id = bq_table_id(
                        telescope.cloud_workspace.project_id,
                        telescope.traffic_partner.bq_dataset_id,
                        telescope.traffic_partner.bq_table_name,
                    )
                    self.assert_table_integrity(table_id, params["bq_rows"])

                    # Test move files to finished
                    ti = env.run_task(telescope.move_files_to_finished.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    for file in release.sftp_files:
                        file_name = os.path.basename(file)
                        in_progress_file = os.path.join(local_sftp_folders.in_progress, file_name)
                        self.assertFalse(os.path.isfile(in_progress_file))

                        finished_file = os.path.join(local_sftp_folders.finished, file_name)
                        self.assertTrue(os.path.isfile(finished_file))

                    # Add_dataset_release_task
                    dataset_releases = get_dataset_releases(
                        dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id
                    )
                    self.assertEqual(len(dataset_releases), 0)
                    ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    dataset_releases = get_dataset_releases(
                        dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id
                    )
                    self.assertEqual(len(dataset_releases), 1)

                    # Test cleanup
                    ti = env.run_task(telescope.cleanup.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)
                    self.assert_cleanup(release.workflow_folder)

    @patch("observatory.platform.airflow.Variable.get")
    def test_gb_transform(self, mock_variable_get):
        """Test sanity check in transform method when transaction date falls outside release month

        :param mock_variable_get: Mock Airflow Variable 'data'
        """
        with CliRunner().isolated_filesystem():
            mock_variable_get.return_value = os.path.join(os.getcwd(), "data")

            # Objects to create release instance
            telescope = GoogleBooksTelescope(
                dag_id="google_books_test",
                cloud_workspace=self.fake_cloud_workspace,
                sftp_root="/",
                sales_partner=partner_from_str("google_books_sales"),
                traffic_partner=partner_from_str("google_books_traffic"),
            )
            fixtures_folder = test_fixtures_folder(workflow_module="google_books_telescope")
            sales_file_path = os.path.join(fixtures_folder, "GoogleSalesTransactionReport_2020_02.csv")
            traffic_file_path = os.path.join(fixtures_folder, "GoogleBooksTrafficReport_2020_02.csv")
            sftp_files = [
                os.path.join(telescope.sftp_folders.in_progress, os.path.basename(sales_file_path)),
                os.path.join(telescope.sftp_folders.in_progress, os.path.basename(traffic_file_path)),
            ]

            # test transaction date inside of release month
            release = GoogleBooksRelease(
                dag_id=telescope.dag_id,
                run_id=random_id(),
                partition_date=pendulum.parse("2020-02-01"),
                sftp_files=sftp_files,
            )
            shutil.copy(sales_file_path, os.path.join(release.download_folder, "google_books_sales.csv"))
            shutil.copy(traffic_file_path, os.path.join(release.download_folder, "google_books_traffic.csv"))
            telescope.transform([release])
            self.assertTrue(os.path.exists(release.transform_sales_path))

            # test transaction date before release month
            release = GoogleBooksRelease(
                dag_id=telescope.dag_id,
                run_id=random_id(),
                partition_date=pendulum.parse("2020-01-31"),
                sftp_files=sftp_files,
            )
            shutil.copy(sales_file_path, os.path.join(release.download_folder, "google_books_sales.csv"))
            shutil.copy(traffic_file_path, os.path.join(release.download_folder, "google_books_traffic.csv"))
            with self.assertRaises(AirflowException):
                telescope.transform([release])

            # test transaction date after release month
            release = GoogleBooksRelease(
                dag_id=telescope.dag_id,
                run_id=random_id(),
                partition_date=pendulum.parse("2020-03-01"),
                sftp_files=sftp_files,
            )
            shutil.copy(sales_file_path, os.path.join(release.download_folder, "google_books_sales.csv"))
            shutil.copy(traffic_file_path, os.path.join(release.download_folder, "google_books_traffic.csv"))
            with self.assertRaises(AirflowException):
                telescope.transform([release])
